# credit for the tooltip to https://stackoverflow.com/questions/3221956/how-do-i-display-tooltips-in-tkinter

import tkinter as tk


class ToolTip:
    """ Create a tooltip for a given widget """

    def __init__(self, widget, text=''):
        """ See help(ToolTip) for more info """
        self.waittime = 500     #miliseconds
        self.wraplength = 180   #pixels
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)
        self.id = None
        self.tw = None

    def enter(self, event=None):
        self.schedule()

    def leave(self, event=None):
        self.unschedule()
        self.hidetip()

    def schedule(self):
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.showtip)

    def unschedule(self):
        id = self.id
        self.id = None
        if id:
            self.widget.after_cancel(id)

    def showtip(self, event=None):
        x = y = 0
        x, y, cx, cy = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = tk.Toplevel(self.widget)
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(self.tw, text=self.text, justify='left',
                       background="#ffffff", relief='solid', borderwidth=1,
                       wraplength = self.wraplength)
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tw
        self.tw= None
        if tw:
            tw.destroy()



def settings(key):
    """ Lay-out settings """
    settings = {

        'lay-out': {
            'width': 600,
            'height': 750,
            'top_height': 30,
            'bottom_height': 80,
        },

        'font': {
            'large': ('Verdana', 12),
            'info': ('Verdana', 10)
        },

        'navigation': {
            'height': 3,
            'width': 20,
            'padx': 5,
            'pady': 5,
        },

        'interactive': {
            'width': 1600,
            'height': 900,
            'design_width': 1000,
            'design_height': 400,
            'info_width': 300,
            'dpi': 150
        },

        'label': {
            'padx': 5,
            'pady': 1,
        },

        'title': {
            'font': ('Verdana', 12, 'bold'),
            'pady': 5
        },

    }

    return settings[key]

def size(to_design, frame, num_concepts):
    """ Returns required frame size based on the structures to design

    Parameters
    ----------
    to_design : dict
        dictionary with the structures to design, key is the structure,
        value is a bool
    frame : {interactive, design}
        for which frame the geometry must be returned
    num_concepts : int
        number of concepts

    Returns
    float, float
        the required width and height of the specified frame
    """
    # width is fixed
    width = 1600

    # check number of structures
    if num_concepts == 1:
        # set frame height
        # check if rubble mound or vertical
        if to_design['RM']['RRM'] or to_design['RM']['CRM']:
            # RubbleMound
            height = 500
            design_height = 400
            design_width = 1000

        elif to_design['C']['RC'] or to_design['C']['CC']:
            # vertical
            height = 600
            design_height = 500
            design_width = 1000

    elif num_concepts == 2:
        # set frame heights
        height = 860
        if to_design['C']['RC'] and to_design['C']['CC']:
            width = 1400
            design_height = 380
            design_width = 800

        else:
            design_height = 380
            design_width = 1000

    if frame == 'interactive':
        return width, height

    elif frame == 'design':
        return design_width, design_height

    else:
        raise NotImplementedError(f'{frame} is not implemented')



import context
import unittest
import numpy as np

from breakwater.core.bishop import Bishop, SlipCircle


class TestBishop(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # example from Soil Mechanics (CTB2310) at TU Delft
        # circle = {'xy': (2.23, 8.20), 'r': 8.5}
        circle = SlipCircle(centre=(2.23, 8.20), r=8.5)
        cls.slip = Bishop(point2=(10.44,6), SlipCircle=circle)
        cls.slip.add_layer(18, 25, phi=5, name='soil')
        cls.slip.compute(num_slices=5)

        # get the slices
        cls.slices = cls.slip.circles[1].slices

    def test_angle(self):
        # slip angles from the example
        # difference in angles is due to rounding intermediate values
        example = [-8.08, 6.14, 20.71, 36.98, 59.5]

        # iterate over the computed values
        for slice, params in self.slices.items():
            self.assertAlmostEqual(
                params['alpha_s'], example[slice]*np.pi/180, 3)

    def test_heights(self):
        # height of slice from the example
        example = [0.75, 1.99, 2.68, 2.69, 1.25]

        # iterate over the computed values
        for slice, params in self.slices.items():
            h = params['h']['soil'][0] - params['h']['soil'][1]
            self.assertAlmostEqual(h, example[slice], 2)

    def test_load(self):
        # loads of each slice from the example
        example = [-1.89, 3.82, 17.04, 29.12, 19.37]

        # iterate over the slices
        for slice, coords in self.slices.items():
            # compute load and strength
            load = self.slip._load(coords['alpha_s'], coords['h'])

            self.assertAlmostEqual(load, example[slice], 2)

    def test_strenght(self):
        # strength of each slice from the example
        # difference in values is again due to intermediate rounding
        # addionally the values of the example for the last two slices
        # do not yield the same result as the given result, the computed
        # result is closer to the result from the Bishop class
        example = [26.77, 28.03, 30.23, 34.33, 46.26]

        # iterate over the slices
        for slice, coords in self.slices.items():
            # compute load and strength
            strength = self.slip._strength(
                alpha_s=coords['alpha_s'], heights=coords['h'], F=1,
                pressure=False)
            self.assertAlmostEqual(strength, example[slice], 2)

    def test_F(self):
        self.assertAlmostEqual(self.slip.circles[1].F, 2.542, 3)


class TestBishopTwoLayers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # example from Soil Mechanics (CTB2310) at TU Delft
        # circle = {'xy': (2.23, 8.20), 'r': 8.5}
        circle = SlipCircle(centre=(2.23, 8.20), r=8.5)

        slip = Bishop(point2=(10.44,6), SlipCircle=circle)
        slip.add_layer(gamma=18, c=25, phi=5, name='layer1', ymin=-20, ymax=0)
        slip.add_layer(gamma=18, c=25, phi=5, name='layer2', ymin=0, ymax=6)
        slip.compute(num_slices=5)

        # set as attribute of the class
        cls.slip = slip

        # get the slices
        cls.slices = cls.slip.circles[1].slices

    def test_angle(self):
        # slip angles from the example
        # difference in angles is due to rounding intermediate values
        example = [-8.08, 6.14, 20.71, 36.98, 59.5]

        # iterate over the computed values
        for slice, params in self.slices.items():
            self.assertAlmostEqual(
                params['alpha_s'], example[slice]*np.pi/180, 3)

    def test_heights(self):
        # height of slice from the example
        example = [0.75, 1.99, 2.68, 2.69, 1.25]

        # iterate over the computed values
        for slice, params in self.slices.items():
            # compute the height of the layer
            h = 0
            for layer, height in params['h'].items():
                h += height[0] - height[1]

            self.assertAlmostEqual(h, example[slice], 2)

    def test_load(self):
        # loads of each slice from the example
        example = [-1.89, 3.82, 17.04, 29.12, 19.37]

        # iterate over the slices
        for slice, coords in self.slices.items():
            # compute load and strength
            load = self.slip._load(coords['alpha_s'], coords['h'])
            self.assertAlmostEqual(load, example[slice], 2)

    def test_strenght(self):
        # strength of each slice from the example
        # difference in values is again due to intermediate rounding
        # addionally the values of the example for the last two slices
        # do not yield the same result as the given result, the computed
        # result is closer to the result from the Bishop class
        example = [26.77, 28.03, 30.23, 34.33, 46.26]

        # iterate over the slices
        for slice, coords in self.slices.items():
            # compute load and strength
            strength = self.slip._strength(
                alpha_s=coords['alpha_s'], heights=coords['h'], F=1,
                pressure=False)
            self.assertAlmostEqual(strength, example[slice], 2)

    def test_F(self):
        self.assertAlmostEqual(self.slip.circles[1].F, 2.542, 3)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from ..utils.wave import dispersion

def toe_stability(Hs, h, ht, Delta, Nod):
    """ Toe stability formula of Van der Meer (1998)

    The armour layer should is supported by a toe, the formula of Van
    der Meer (1998) computes the minimal required nominal diameter of
    the stones necessary. The formula is given as:

    .. math::
       \\frac{H_{s}}{\\Delta D_{n50}}=\\left(6.2 \\frac{h_{t}}{h}+2
       \\right) N_{o d}^{0.15}

    The formula is based on experiments and the range of validity of
    the parameters can be seen in the table below.

    +---------------------------------+---------+-----------+
    | Parameter                       | Symbol  | Range     |
    +=================================+=========+===========+
    | toe depth over water depth      | ht/h    | 0.4 - 0.9 |
    +---------------------------------+---------+-----------+
    | toe depth over nominal diameter | ht/Dn50 |   3 - 25  |
    +---------------------------------+---------+-----------+

    Parameters
    ----------
    Hs : float
        The significant wave height of the incident waves [m]
    h : float
        The water depth in front of the toe [m]
    ht : float
        The water depth on top of the toe [m]
    Delta : float
        Relative buoyant density [-]
    Nod : float
        Damage number [-]

    Returns
    -------
    Dn50 : float
        Nominal diameter of the armourstones in the toe [m]
    """
    Dn50 = Hs/(Delta*(2 + 6.2*(ht/h)**2.7)*Nod**0.15)
    return Dn50


def toe_berm_stability(Hs, T, d, Bm, Delta, beta=0, alpha_s=0.45):
    """ Berm protection to caisson or vertical wall breakwaters

    Compute the nominal diameter of the armourstone on the berm of a
    caisson or vertical wall breakwaters with the modified Tanimoto
    formula from Takahasi (2002).

    .. math::
       D_{n50} = \\frac{H_{s}}{\\Delta N_{s}}

    in which:

    .. math::
       N_{s}=\\max \\left\\{1.8,\\left(1.3 \\frac{1-\\kappa}{\\kappa^{1 / 3}}
       \\frac{h^{\\prime}}{H_{s}}+1.8 \\exp \\left[-1.5 \\frac{
       (1-\\kappa)^{2}}{\\kappa^{1 / 3}} \\frac{h^{\\prime}}{H_{s}}
       \\right]\\right)\\right\\}

    where:

    .. math::
       \\kappa=\\frac{4 \\pi h^{\\prime} / L^{\\prime}}{\\sinh \\left(4
       \\pi h^{\\prime} / L^{\\prime}\\right)} \\kappa_{2}

    and:

    .. math::
       \\kappa_{2}=\\max \\left\{\\alpha_{S} \\sin ^{2} \\beta \\cos ^{2}
       \\left(\\frac{2 \\pi x}{L^{\\prime}} \\cos \\beta\\right)\\right.,
       \\left.\\cos ^{2} \\beta \\sin ^{2}\\left(\\frac{2 \\pi x}
       {L^{\\prime}} \\cos \\beta\\right)\\right\\}


    Parameters
    ----------
    Hs : float
        mean of the highest 1/3 of the wave heights [m]
    T : float
        wave period (s)
    d : float
        water depth at which the armour is placed [m]
    Bm : float
        width of the berm [m]
    Delta : float
        Relative buoyant density [-]
    beta : float, optional, default: 0
        angle between direction of wave approach and a line normal to
        the breakwater [rad]
    alpha_s : float, optional, default: 0.45
        factor the include the effect of the slope, the value of 0.45 is
        given by measures data (Goda, 2000)

    Returns
    -------
    Dn50 : float
        Nominal diameter of the armourstones on the berm [m]
    """
    L = dispersion(T=T, h=d)
    kh = 2*np.pi*d/L

    kappa2 = max([alpha_s*np.sin(beta)**2 * np.cos(2*np.pi*Bm*np.cos(beta)/L)**2,
                  np.cos(beta)**2 * np.sin(2*np.pi*Bm*np.cos(beta)/L)**2])

    kappa = 2*kh/(np.sinh(2*kh)) * kappa2

    a = (1-kappa)/kappa**(1/3)
    Ns = (1.3 * a * d/Hs + 1.8 * np.exp(-1.5*d*a*(1-kappa)/Hs))
    Ns = max([1.8, Ns])

    Dn50 = Hs/(Delta*Ns)

    return Dn50



import tkinter as tk
from webbrowser import open_new

from .settings import settings

def footer(frame):
    """ Add footer

    Function to add a footer with text and a link to a LinkedIn page

    Parameters
    ----------
    frame : tk.Frame
        frame to which the footer must be added
    """
    # add footer
    footer = tk.Label(
        frame, text='Developed by S. Winkel. Breakwater is licensed under CC BY-NC-SA 4.0', bd=1, relief=tk.SUNKEN,
        anchor='w')
    footer.pack(side=tk.BOTTOM, fill=tk.X)
    footer.bind(
        "<Button-1>",
        lambda e: open_new('https://www.linkedin.com/in/sander-winkel/'))



import numpy as np

from ..utils.exceptions import user_warning, NotSupportedError

def gamma_f(
        armour_layer, xi_m_min_1, layers=None, permeability=None,
        placement=None):
    """ Influence factor for the permeability and roughness of the slope

    Computes the influence factor on roughness with table 6.2 from
    EurOtop (2018). These values are derived for 2.8
    :math:`\\leq \\xi_{m-1.0} \\leq` 4.5, in case of larger surf-similarity
    parameters the influence factor for roughness is increased using
    equation 6.7 from EurOtop (2018)

    +--------------------------------------+-------------------+
    | Type of armour layer                 | :math:`\\gamma_f`  |
    +======================================+===================+
    | Smooth impermeable surface           |       1.00        |
    +--------------------------------------+-------------------+
    | Rock (1 layer, impermeable core)     |       0.60        |
    +--------------------------------------+-------------------+
    | Rock (1 layer, permeable core)       |       0.45        |
    +--------------------------------------+-------------------+
    | Rock (2 layers, impermeable core)    |       0.55        |
    +--------------------------------------+-------------------+
    | Rock (2 layers, permeable core)      |       0.40        |
    +--------------------------------------+-------------------+
    | Cubes (1 layer, flat positioning)    |       0.49        |
    +--------------------------------------+-------------------+
    | Cubes (2 layers, random positioning) |       0.47        |
    +--------------------------------------+-------------------+
    | Antifers                             |       0.50        |
    +--------------------------------------+-------------------+
    | HARO                                 |       0.47        |
    +--------------------------------------+-------------------+
    | Tetrapods                            |       0.38        |
    +--------------------------------------+-------------------+
    | Dolos                                |       0.43        |
    +--------------------------------------+-------------------+
    | Accropode I                          |       0.46        |
    +--------------------------------------+-------------------+
    | Xbloc, CoreLoc, Accropode II         |       0.44        |
    +--------------------------------------+-------------------+
    | XblocPlus                            |       0.45        |
    +--------------------------------------+-------------------+
    | Cubipods (1 layer)                   |       0.49        |
    +--------------------------------------+-------------------+
    | Cubipods (2 layers)                  |       0.47        |
    +--------------------------------------+-------------------+


    Parameters
    ----------
    armour_layer : str
        name of the material of the armour layer, supported materials:
        Smooth, Rock, Cubes, Antifers, HARO, Tetrapods, Dolos,
        Accropode I, Xbloc, XblocPlus, CoreLoc, Accropode II, Cubipods
    xi_m_min_1 : float
        the surf-similarity parameter [-]
    layers : {1, 2}, optional, default: None
        number of layers in the armour layer, required if the armour
        layer is made out of: Rock, Cubes or Cubipods
    permeability : {'permeable', 'impermeable'}, optional, default: None
        permeability of the armour layer, required if the armour layer
        is made out of Rock
    placement : {'flat', 'random'}, optional, default: None
        placement of the armour layer, required if the armour layer is
        made out of Cubes

    Returns
    -------
    gamma_f : float
        the influence factor for roughness

    Raises
    ------
    Keyerror
        If the armour layer is not in table 6.2 from EurOtop (2018)
    """
    table = {'Smooth': 1,
             'Rock - 1 - permeable': 0.6,
             'Rock - 1 - impermeable': 0.45,
             'Rock - 2 - permeable': 0.55,
             'Rock - 2 - impermeable': 0.40,
             'Cubes - 1 - flat': 0.49,
             'Cubes - 2 - random': 0.47,
             'Antifers': 0.5,
             'HARO': 0.47,
             'Tetrapods': 0.38,
             'Dolos': 0.43,
             'Accropode I': 0.46,
             'Xbloc': 0.44,
             'XblocPlus': 0.45,
             'CoreLoc': 0.44,
             'Accropode II': 0.44,
             'Cubipods - 1': 0.49,
             'Cubipods - 2': 0.47}
    if armour_layer == 'Smooth':
        key = 'Smooth'
    elif armour_layer == 'Rock':
        key = f'Rock - {layers} - {permeability}'
    elif armour_layer == 'Cubipods':
        key = f'Cubipods - {layers}'
    elif armour_layer == 'Cubes':
        key = f'Cubes - {layers} - {placement}'
    else:
        key = f'{armour_layer}'

    # check if the armour layer (key) is in the table
    if key in table.keys():
        # get the gamma_f value
        gamma_f = table[key]
    else:
        # raise error as the given armour layer is not supported
        raise NotSupportedError(
            ('EurOtop (2018) does not support a roughness factor for '
             f'{armour_layer}'))

    if xi_m_min_1 > 5:
        gamma_f = gamma_f + (xi_m_min_1-5)*(1-gamma_f)/5
    elif xi_m_min_1 < 2.8:
        user_warning(
            (f'xi out of range in gamma_f, {np.round(xi_m_min_1,2)} < 2.8. '
             f'Continued with gamma_f = {gamma_f}'))
    return gamma_f

def gamma_beta(beta):
    """ Influence factor for oblique wave attack

    Computes the influence factor for oblique wave attack with equation
    5.29 from EurOtop (2018).

    .. math::
       \\gamma_{\\beta} = 1 - 0.0033 \\mid \\beta \\mid

    with a maximum of :math:`\\gamma_{\\beta} = 0.736` for
    :math:`\\mid \\beta \\mid > 80`

    Parameters
    ----------
    beta : float
        the angle of wave attack [rad]

    Returns
    -------
    gamma_beta : float
        the influence factor for oblique wave attack
    """
    beta_deg = beta*180/np.pi

    if beta_deg <= 80:
        return 1 - 0.0033 * np.abs(beta_deg)
    else:
        return 0.736

def rubble_mound(
        Hm0, q, xi_m_min_1, alpha, beta, gamma_b, gamma_v, gam_star,
        armour_layer, layers=1, permeability='permeable', safety=1,
        Gc=None, Dn50=None, limit=True):
    """ Compute the crest freeboard of a rubble mound breakwater

    Computes the crest freeboard of a rubble mound breakwater using
    equation 5.10 and 5.12 from EurOtop (2018).

    .. math::
       \\frac{q}{\\sqrt{g \\cdot H_{m 0}^{3}}}=\\frac{0.023}{\\sqrt{\\tan
       \\alpha}}\\gamma_{b} \\cdot \\xi_{m-1,0} \\cdot \\exp \\left[-
       \\left(2.7 \\frac{R_{c}}{\\xi_{m-1,0} \\cdot H_{m 0} \\cdot
       \\gamma_{b} \\cdot \\gamma_{f} \\cdot \\gamma_{\\beta} \\cdot
       \\gamma_{v}}\\right)^{1.3}\\right]

    with a maximum of

    .. math::
       \\frac{q}{\\sqrt{g \\cdot H_{m 0}^{3}}}=0.09 \\cdot \\exp \\left
       [-\\left(1.5 \\frac{R_{c}}{H_{m 0} \\cdot \\gamma_{f} \\cdot
       \\gamma_{\\beta} \\cdot \\gamma^{*}}\\right)^{1.3}\\right]

    The reliability of the first equation is given by
    :math:`\\sigma` (0.023) = 0.003 and :math:`\\sigma` (2.7) = 0.20,
    and for the second equation by :math:`\\sigma` (0.09) = 0.0135 and
    :math:`\\sigma` (1.5) = 0.15.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    xi_m_min_1 : float
        :math:`\\xi_{m-1.0}`, the surf-similarity parameter computed
        with the energy wave period :math:`T_{m-1.0}` [-]
    alpha : float
        Angle of the front slope [rad]
    beta : float
        the angle of wave attack [rad]
    gamma_b : float
        influence factor for a berm [-]
    gamma_v : float
        influence factor for a vertical wall [-]
    gamma_star : float
        overall influence factor for a storm wall on slope or promenade [-]
    armour_layer : str
        name of the material of the armour layer, supported materials:
        Smooth, Rock, Cubes, Antifers, HARO, Tetrapods, Dolose,
        Accropode I, Xbloc, CoreLoc, Accropode II, Cubipods
    layers : {1, 2}, default: 1
        number of layers in the armour layer, required if the armour
        layer is made out of: Rock, Cubes or Cubipods
    permeability : {'permeable', 'impermeable'}, default: 'permeable'
        permeability of the armour layer, required if the armour layer
        is made out of Rock
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    Gc : float, optional, default: None
        Effect of armoured crest width [m]
    Dn50 : float, optional, default: None
        nominal diameter of the armour units on the crest [m]
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    gam_f = gamma_f(
        armour_layer=armour_layer, xi_m_min_1=xi_m_min_1, layers=layers,
        permeability=permeability)
    gam_beta = gamma_beta(beta)

    constant_a = 0.023 + safety*0.003
    constant_b = 2.7 - safety*0.2
    constant_c = 0.09 + safety*0.0135
    constant_d = 1.5 - safety*0.15

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # in case of a wide crest the overtopping in the equation can be increased
    if Gc != None and Gc > 3*Dn50:
        Cr = 3.06 * np.exp(-1.5*Gc/Hm0)
        q = q / min(Cr, 1)

    q = q/1000

    # compute arguments of the ln (to check if arg < 1)
    arg = (q * np.sqrt(np.tan(alpha))
           / (constant_a*gamma_b*xi_m_min_1*np.sqrt(g*Hm0**3)))
    arg_upper = q/(constant_c*np.sqrt(g*Hm0**3))

    if arg <= 1:
        Rc = ((xi_m_min_1*Hm0*gamma_b*gam_f*gam_beta*gamma_v/constant_b)
              * (-np.log(arg))**(1/1.3))
    else:
        Rc = 0
        user_warning('Encountered negative freeboard')

    if arg_upper <= 1: # handle for RuntimeWarning
        Rc_upper = ((Hm0*gam_f*gam_beta*gam_star/constant_d)
                    * (-np.log(arg_upper))**(1/1.3))
    else:
        Rc_upper = 0
        user_warning('Encountered negative freeboard in upper bound')

    Rc = min(Rc, Rc_upper)

    return Rc

def vertical_deep(Hm0, q, safety=1, limit=True):
    """ Rc if the foreshore does not have an influence

    Compute the crest freeboard for a vertical or composite vertical
    wall if the foreshore does not have an influence. Implementation
    of equation 7.1 from EurOtop (2018).

    .. math::
       \\frac{q}{\\sqrt{g \\cdot H_{m 0}^{3}}}=0.047 \\cdot \\exp \\left[
       -\\left(2.35 \\frac{R_{c}}{H_{m 0}}\\right)^{1.3}\\right]

    The reliability of the equation is given by
    :math:`\\sigma` (0.047) = 0.007 and :math:`\\sigma` (2.35) = 0.23.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant_a = 0.047 + safety*0.007
    constant_b = 2.35 - safety*0.23

    Rc = (1/constant_b*Hm0
          * (-np.log(q/(constant_a*np.sqrt(g*Hm0**3))))**(1/1.3))
    return Rc

def vertical_no_breaking(Hm0, q, safety=1, limit=True):
    """ Rc if no possibility for breaking waves

    Compute the crest freeboard for a vertical or composite vertical
    wall if there are no breaking waves. Implementation of equation 7.5
    from EurOtop (2018)

    .. math::
       \\frac{q}{\\sqrt{g H_{m 0}^{3}}}=0.05 \\exp \\left(-2.78
       \\frac{R_{c}}{H_{m 0}}\\right)

    The reliability of the equation is given by
    :math:`\\sigma` (0.05) = 0.012 and :math:`\\sigma` (2.78) = 0.17.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant_a = 0.05 + safety*0.012
    constant_b = 2.78 - safety*0.17

    Rc = 1/constant_b*Hm0 * -np.log(q/(constant_a*np.sqrt(g*Hm0**3)))
    return Rc

def vertical_normal(Hm0, q, h, s_m_min_1, safety=1, limit=True):
    """ Rc for a vertical wall if normal freeboard and breaking waves

    Compute the crest freeboard for a vertical wall if there is a
    possibility for breaking waves, but the freeboard is not low.
    Implementation of equation 7.8 from EurOtop (2018)

    .. math::
       \\frac{q}{\\sqrt{g H_{m 0}^{3}}}=0.0014\\left(\\frac{H_{m 0}}
       {h s_{m-1,0}}\\right)^{0.5}\\left(\\frac{R_{c}}{H_{m 0}}
       \\right)^{-3}

    The reliability of the equation is given by
    :math:`\\sigma` (0.0014) = 0.0006

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    h : float
        water depth in front of the toe of the structure [m]
    s_m_min_1 : float
        :math:`s_{m-1.0}`, wave steepness with the spectral wave length
        (:math:`L_{m-1.0}`) [-]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant = 0.0014 + safety*0.0006

    Rc = Hm0 * (q/(np.sqrt(Hm0*g*Hm0**3/(h*s_m_min_1))*constant))**(1/-3)
    return Rc

def vertical_low(Hm0, q, h, s_m_min_1, safety=1, limit=True):
    """ Rc for a vertical wall if low freeboard and breaking waves

    Compute the crest freeboard for a vertical wall if there is a
    possibility for breaking waves, and the freeboard is low.
    Implementation of equation 7.7 from EurOtop (2018)

    .. math::
       \\frac{q}{\\sqrt{g H_{m 0}^{3}}}=0.011\\left(\\frac{H_{m 0}}{h
       s_{m-1,0}}\\right)^{0.5} \\exp \\left(-2.2 \\frac{R_{c}}{H_{m 0}}
       \\right)

    The reliability of the equation is given by
    :math:`\\sigma` (0.011) = 0.0045

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    h : float
        water depth in front of the toe of the structure [m]
    s_m_min_1 : float
        :math:`s_{m-1.0}`, wave steepness with the spectral wave length
        (:math:`L_{m-1.0}`) [-]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant = 0.011 + safety*0.0045

    Rc = Hm0/-2.2 * np.log(q/(np.sqrt(Hm0*g*Hm0**3/(h*s_m_min_1))*constant))
    return Rc

def composite_normal(Hm0, q, h, d, s_m_min_1, safety=1, limit=True):
    """ Rc for a composite wall if normal freeboard and breaking waves

    Compute the crest freeboard for a composite vertical wall if there
    is a possibility for breaking waves, but the freeboard is not low.
    Implementation of equation 7.14 from EurOtop (2018)

    .. math::
       \\frac{q}{\\sqrt{g H_{m 0}^{3}}}=1.3\\left(\\frac{d}{h}\\right)^{0.5}
       0.0014\\left(\\frac{H_{m 0}}{h s_{m-1,0}}\\right)^{0.5}\\left(
       \\frac{R_{c}}{H_{m 0}}\\right)^{-3}

    The reliability of the equation is given by
    :math:`\\sigma` (0.0014) = 0.0006.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    h : float
        water depth in front of the toe of the structure [m]
    d : float
        water depth above the toe of the structure [m]
    s_m_min_1 : float
        :math:`s_{m-1.0}`, wave steepness with the spectral wave length
        (:math:`L_{m-1.0}`) [-]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant = 0.0014 + safety*0.0006

    Rc = Hm0 * (q/(np.sqrt(d*g*Hm0**4/(h**2*s_m_min_1))*1.3*constant))**(1/-3)
    return Rc

def composite_low(Hm0, q, h, d, s_m_min_1, safety=1, limit=True):
    """ Rc for a composite wall if low freeboard and breaking waves

    Compute the crest freeboard for a composite vertical wall if there
    is a possibility for breaking waves, and the freeboard is low.
    Implementation of equation 7.15 from EurOtop (2018)

    .. math::
       \\frac{q}{\\sqrt{g H_{m 0}^{3}}}=1.3\\left(\\frac{d}{h}
       \\right)^{0.5} 0.011\\left(\\frac{H_{m 0}}{h s_{m-1,0}}
       \\right)^{0.5} \\exp \\left(-2.2 \\frac{R_{c}}{H_{m 0}}\\right)

    The reliability of the equation is given by
    :math:`\\sigma` (0.011) = 0.0045.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    h : float
        water depth in front of the toe of the structure [m]
    d : float
        water depth above the toe of the structure [m]
    s_m_min_1 : float
        :math:`s_{m-1.0}`, wave steepness with the spectral wave length
        (:math:`L_{m-1.0}`) [-]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    g = 9.81

    # check given overtop discharge against EurOtop limit for zero discharge
    limit_zero_q = np.sqrt(g*Hm0**3) * 0.01
    if limit_zero_q > q:
        msg = (f'Given value of q = {q}, is below EurOtop limit for zero '
               f'discharge, which is {np.round(limit_zero_q, 3)}')

        if limit:
            msg += f'. Using {np.round(limit_zero_q, 3)} instead as limit=True'
            q = limit_zero_q

        user_warning(msg)

    # convert q from l to m^3
    q = q/1000

    constant = 0.011 + safety*0.0045
    Rc = Hm0/-2.2*np.log(q/(np.sqrt(d*g*Hm0**4/(h**2*s_m_min_1))*1.3*constant))
    return Rc

def vertical(
        Hm0, q, h, d, L_m_min_1, s_m_min_1, safety=1, logger=None, limit=True):
    """ Compute crest freeboard for vertical and composite vertical walls

    Compute the crest freeboard, Rc, of a vertical or composite vertical
    wall for a given mean overtop discharge. The function is an
    implementation of the decision chart, figure 7.2, from EurOtop
    (2018). The function determines if the input classifies as a
    vertical or composite vertical wall, if breaking is possible and if
    the structure has a low freeboard. Based on the classification the
    function calls the corresponding function and computes the freeboard.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    q : float
        mean overtopping discharge per meter structure width [l/s per m]
    h : float
        water depth in front of the toe of the structure [m]
    d : float
        water depth above the toe of the structure [m]
    L_m_min_1 : float
        :math:`L_{m-1.0}`, spectral wave length in deep water [m]
    s_m_min_1 : float
        :math:`s_{m-1.0}`, wave steepness with the spectral wave length
        (:math:`L_{m-1.0}`) [-]
    safety : float, optional, default: 1
        safety factor of the design, positive values increase the safety
        of the design by increasing the mean value of the model constants
        with the number of standard deviations specified. In accordance
        with the recommendation from EurOtop (2018), the default value is
        set to 1 standard deviation.
    logger : dict, optional, default: None
        dict to log messages, must have keys 'INFO' and 'WARNINGS'
    limit : bool, optional, default: True
        If True, the discharge will be set to the limit for zero
        discharge in case the given discharge is below this limit. If
        False, the discharge will not be changed.

    Returns
    -------
    Rc : float
        the crest freeboard of the structure [m]
    """
    if h/Hm0 > 4:
        Rc = vertical_deep(Hm0, q, safety, limit)
        log = 'overtopping for vertical in deep water'
    else:
        if d/h > 0.6:  # check if vertical wall
            if h**2/(Hm0*L_m_min_1) < 0.23:  # check if waves are breaking
                Rc = vertical_normal(Hm0, q, h, s_m_min_1, safety, limit)
                log = 'overtopping for vertical with normal freeboard'
                if Rc/Hm0 < 1.35:  # check if freeboard is low
                    Rc = vertical_low(Hm0, q, h, s_m_min_1, safety, limit)
                    log = 'overtopping for vertical with low freeboard'
            else:
                Rc = vertical_no_breaking(Hm0, q, safety, limit)
                log = 'overtopping for vertical with no breaking waves'
        else:  # not a vertical wall, treat as composite wall
            if h*d/(Hm0*L_m_min_1) < 0.65:  # check if waves are breaking
                Rc = composite_normal(Hm0, q, h, d, s_m_min_1, safety, limit)
                log = 'overtopping for composite with normal freeboard'
                if Rc/Hm0 < 1.35:  # check if freeboard is low
                    Rc = composite_low(Hm0, q, h, d, s_m_min_1, safety, limit)
                    log = 'overtopping for composite with low freeboard'
            else:
                Rc = vertical_no_breaking(Hm0, q, safety, limit)
                log = 'overtopping for composite with no breaking waves'
    if logger != None:
        logger['INFO'].append(log)

    return Rc



from matplotlib.widgets import Slider

from .utils._kwarg_validator import _RM_vkwargs, _C_vkwargs
from .app.main import BreakwaterDesign
from .utils.exceptions import InputError, NotSupportedError
from .utils.cost import _process_cost


def interactive_design(
        structure, LimitState, Grading, ArmourUnit=None, BermMaterial=None,
        cost=None, Soil=None, display_warnings=True):
    """ Interactive Design Application

    Application to interactively design breakwaters. All parameters can
    be specified in the application, together with varying parameters.
    On the Interactive Design Page it is then possible to change the
    value of a varying parameter with a slider. The design can then be
    updated and a cross-section of the new design is shown.

    .. note
       The maximum number of structures that can be specified is 2.

    Parameters
    ----------
    structure : {'RRM', 'CRM', 'RC', 'CC'}
        structure for which conceptual designs must be generated. RRM
        for a rubble mound with rock as armour layer, CRM for a rubble
        mound with concrete armour units as armour layer, RC for a
        vertical (composite) breakwater with rock as armour layer for
        the foundation and CC for a vertical (composite) breakwater with
        concrete armour units as armour layer for the foundation.
    LimitState : :py:class:`LimitState` or list of :py:class:`LimitState`
        ULS, SLS or another limit state defined with
        :py:class:`LimitState`
    Grading : :py:class:`RockGrading`
        standard rock grading defined in the NEN-EN 13383-1 or a user
        defined rock grading. Required for all parameters
    ArmourUnit : obj, optional, default: None
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`. This
        argument is used for RRM.
    BermMaterial : obj, optional, default: None
        armour unit class which inherits from :py:class:`ConcreteArmour`,
        for instance :py:class:`Xbloc` or :py:class:`XblocPlus`. This
        argument is used for CC.
    Soil : :py:class:`Soil`, optional, default: None
        by default Soil is None, which means that the geotechnical checks
        are not performed. By specifying a Soil object, the geotechnical
        checks are automatically performed. Optional argument for RC and
        CC.
    cost : dict, optional, default: None
        by default no cost are computed, when the cost must be computed
        all relevant cost must be included. Optional keys of the dict
        are: core_price for the price of the core material, unit_price
        for the price of armour units,  concrete_price for the price of
        concrete, fill_price for the price of the fill material of the
        caisson. If not included in the prices it is also possible to
        specify transport_price for the price of transporting rocks,
        dry_dock for the rent of a dry dock, which is divided through
        the length of the breakwater.
    display_warnings : bool, optional, default: True
        if warnings must be displayed when designing
    """
    # convert the input of structure to a list
    if isinstance(structure, list):
        # must be a list so no change
        structure = structure
    elif isinstance(structure, str):
        # convert single input to list
        structure = [structure]

    # convert single LimitState to list if needed
    if isinstance(LimitState, list):
        LimitStates = LimitState
    else:
        LimitStates = [LimitState]

    # create dict of the python input for the application
    python_input = {
        'LimitState': LimitStates, 'Grading': Grading, 'Soil': Soil}

    # process python input
    if 'CRM' in structure:
        if ArmourUnit is None:
            raise InputError(f'ArmourUnit is a required argument for CRM')
        else:
            python_input['ArmourUnit'] = ArmourUnit

    if 'CC' in structure:
        if BermMaterial is None:
            raise InputError(f'BermMaterial is a required argument for CC')
        else:
            python_input['BermMaterial'] = BermMaterial

    # dict with the structures to design
    to_design = {
        'RM': {'num': 0, 'RRM': False, 'CRM': False},
        'C': {'num': 0, 'RC': False, 'CC': False}
        }

    # dict to store the vkwargs in
    vkwargs = {}

    # check structure type
    if 'RRM' in structure and 'CRM' in structure:
        vkwargs.update(_RM_vkwargs(type='both'))
        to_design['RM']['RRM'] = True
        to_design['RM']['CRM'] = True
        to_design['RM']['num'] = 2

    elif 'RRM' in structure:
        vkwargs.update(_RM_vkwargs(type='Rock'))
        to_design['RM']['RRM'] = True
        to_design['RM']['num'] = 1

    elif 'CRM' in structure:
        vkwargs.update(_RM_vkwargs(type='ArmourUnit'))
        to_design['RM']['CRM'] = True
        to_design['RM']['num'] = 1

    if 'RC' in structure and 'CC' in structure:
        vkwargs.update(_C_vkwargs(type='both'))
        to_design['C']['RC'] = True
        to_design['C']['CC'] = True
        to_design['C']['num'] = 2

    elif 'RC' in structure:
        vkwargs.update(_C_vkwargs(type='Rock'))
        to_design['C']['RC'] = True
        to_design['C']['num'] = 1

    elif 'CC' in structure:
        vkwargs.update(_C_vkwargs(type='ArmourUnit'))
        to_design['C']['CC'] = True
        to_design['C']['num'] = 1

    # check number of structures
    # due to fixed lay out the maximum number is 2
    total_num = to_design['RM']['num'] + to_design['C']['num']
    if total_num > 2:
        # to many structures
        raise NotSupportedError(
            ('Too many structures have been specified, currently only 2 '
             'structures can be designed at the same time'))

    # delete python input from parameters as these are given in Python
    for parameter in python_input.keys():
        if parameter in vkwargs.keys():
            del vkwargs[parameter]

    # process the cost
    cost = _process_cost(structure, cost, Grading, validate=False)

    # start the app
    app = BreakwaterDesign(
        vkwargs, python_input, to_design, cost, display_warnings)
    app.mainloop()



import tkinter as tk
import os

from .settings import settings, size
from .input import ParameterInput, VaryingInput
from .interactive import InteractiveDesign
from .footer import footer


class BreakwaterDesign(tk.Tk):
    """ Tkinter app for interactive design

    Tkinter application to design breakwaters

    Parameters
    ----------
    vkwargs : dict
        dictionary of the valid arguments for the specified structures
    input : dict
        dictionary with the Python input, this is at least a LimitState
        and Grading.
    to_design : dict
        dictionary with the structures to design, key is the structure,
        value is a bool
    cost : dict
        dictionary with the price of concrete and sand
    display_warnings : bool
        True if warnings must be displayed when designing, False if not
    """

    def __init__(
            self, vkwargs, input, to_design, cost, display_warnings,
            *args, **kwargs):
        """ See help(BreakwaterDesign) for more info """
        # initialize tkinter
        tk.Tk.__init__(self, *args, **kwargs)

        # set attribute
        self.vkwargs = vkwargs
        self.parameters = input
        self.to_design = to_design
        self.cost = cost
        self.display_warnings = display_warnings

        # set attribute for the previous page of InteractiveDesign
        self.previous = 'ParameterInput'

        # determine number of concepts and set as attribute
        self.num_concepts = to_design['RM']['num'] + to_design['C']['num']

        # set icon of the app
        file = os.path.dirname(os.path.abspath(__file__))
        icon_path = os.path.join(file, 'bw-icon.ico')
        try:
            self.iconbitmap(icon_path)
        except:
            print(f'Could not load the icon from {icon_path}')

        # create a containter
        container = tk.Frame(self)

        # format main lay-out
        container.pack(side='top', fill='both', expand=True)
        container.rowconfigure(0, weight=1)
        container.columnconfigure(0, weight=1)

        # add title
        self.title('Interactive Breakwater Design')

        # create dict to store frames in
        self.frames = {}

        # add frames
        for Frame in (StartPage, ParameterInput, InteractiveDesign, VaryingInput):
            frame = Frame(container, self)
            self.frames[Frame] = frame
            frame.grid(row=0, column=0, sticky='nsew')

        # set and show home page (=start page)
        self.home_page = StartPage
        self.home()

    def show_frame(self, controller, default=True):
        """ Load a new frame

        Parameters
        ----------
        controller : tk.Frame
            The frame to show
        default : bool, optional, default: True
            True if the default geometry settings must be used, False
            if the adapted geometry settings must be showed
        """
        if default:
            # get width and height
            width = settings('lay-out')['width']
            height = settings('lay-out')['height']
        else:
            # position app in the center of the screen
            width, height = size(
                self.to_design, 'interactive', self.num_concepts)

        ws = self.winfo_screenwidth()
        hs = self.winfo_screenheight()
        x = (ws/2) - (width/2)
        y = (hs/2) - (height/2)
        self.geometry(f'{width}x{height}+{round(x)}+{round(y)}')

        # get frame
        frame = self.frames[controller]

        # display the frame
        frame.tkraise()

    def home(self):
        """ Go to the home page """
        self.show_frame(self.home_page)

    def back(self, page):
        """ Go back to the previous page

        Method to go back to the previous page from InteractiveDesign

        Parameters
        ----------
        page : str
            name of the page
        """
        if page == 'VaryingInput':
            self.show_frame(VaryingInput)

        elif page == 'ParameterInput':
            self.show_frame(ParameterInput)

        else:
            raise NotImplementedError(f'{page} is an invalid page name')


class StartPage(tk.Frame):
    """ Start page of the app """

    def __init__(self, parent, controller):
        """ see help(StartPage) for more info """
        # create the frame
        tk.Frame.__init__(self, parent)

        # get width and height
        width = settings('lay-out')['width']
        height = settings('lay-out')['height']

        # set height of frames
        top_height = settings('lay-out')['top_height']
        bottom_height = settings('lay-out')['bottom_height']
        center_height = height - top_height - bottom_height - 50

        # create lay out
        top_frame = tk.Frame(self, width=width, height=top_height)
        top_frame.pack_propagate(False)
        center_frame = tk.Frame(self, width=width, height=center_height)
        center_frame.pack_propagate(False)
        bottom_frame = tk.Frame(self, width=width, height=bottom_height)
        bottom_frame.pack_propagate(False)

        top_frame.pack(side=tk.TOP, pady=5)
        center_frame.pack()
        bottom_frame.pack(side=tk.BOTTOM)

        # add title of the page
        label = tk.Label(
            center_frame,
            text='Welcome to the Interactive Design Application of breakwater',
            **settings('title'))
        label.place(relx=0.5, rely=0.25, anchor=tk.CENTER)

        # welcome message
        label = tk.Label(
            center_frame,
            text=(''
            'All parameters required for the design can be given on the '
            'next page. On this page you can also specify \n which parameters'
            ' are allowed to vary, for each of these parameters a slider is '
            'generated. The minimum and \n maximum values for these sliders '
            'can be specified on the Varying Input page, which is '
            'accessed by checking \n the checkbox behind the parameter and '
            'pressing Start Design. \n'
            '\n'
            'When all required input has been given the Interactive Design '
            'page can be accessed, on this page a \n cross-section of the '
            'specified breakwater structure(s) is presented together with '
            'sliders if parameters \n are allowed to vary. The design can be '
            'updated by changing the sliders and pressing the \n Update '
            'Design button. A new cross-section is then plotted, with the '
            'previous design in grey.')
        )
        label.place(relx=0.5, rely=0.45, anchor=tk.CENTER)

        # add the footer
        footer(bottom_frame)

        # add next button
        nav_settings = settings('navigation')

        next_button = tk.Button(
            bottom_frame, text='Next', height=nav_settings['height'],
            width=nav_settings['width'],
            command=lambda: controller.show_frame(ParameterInput))
        next_button.pack(
            side=tk.RIGHT, padx=nav_settings['padx'], pady=nav_settings['pady'])



import warnings

# define errors
class InputError(Exception):
    """ Exception raised for errors in the input

    Attributes
    ----------
    message
        explanation of the error
    """
    __module__ = Exception.__module__

    def __init__(self, message):
        self.message = message


class NotSupportedError(Exception):
    """ Exception raised when input is not supported

    Attributes
    ----------
    message
        explanation of the error
    """
    __module__ = Exception.__module__

    def __init__(self, message):
        self.message = message


class RockGradingError(Exception):
    """ Exception raised if there is an error in the Rock Grading

    Attributes
    ----------
    message
        explanation of the error
    """
    __module__ = Exception.__module__

    def __init__(self, message):
        self.message = message


class ArmourUnitsError(Exception):
    """ Exception raised if there is an error in the Armour Units

    Attributes
    ----------
    message
        explanation of the error
    """
    __module__ = Exception.__module__

    def __init__(self, message):
        self.message = message


# define warnings
class LimitStateWarning(Warning):
    """ Warning in the LimitState """

    def __init__(self, msg):
        self.message = msg


# monkeypatch warning format
def _custom_formatwarning(msg, category, *args, **kwargs):
    # ignore everything except the message
    return f'{category.__name__}: {msg} \n'

def user_warning(msg):
    """ show user warning """
    warnings.formatwarning = _custom_formatwarning
    warnings.warn(msg, category=UserWarning)

def limitstate_warning(msg):
    """ Show warning in the LimitState """
    warnings.formatwarning = _custom_formatwarning
    warnings.warn(msg, category=LimitStateWarning)

def no_warnings():
    """ ignore warnings """
    warnings.filterwarnings('ignore')



import context
import unittest

from breakwater.core.soil import Soil


class TestSoil(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.soil = Soil(c=20, phi=15, gamma=19, n=0.4)

    def test_brinch_hansen(self):
        # example from CTB2310 Soil Mechanics at the TU Delft
        # no inclination factors
        p = self.soil.brinch_hansen(p=0, t=0, B=10, L=10, q=19, sat=False)
        self.assertAlmostEqual(p, 462.51, 2)

        # with inclination factors
        p = self.soil.brinch_hansen(t=30, p=175, B=10, L=10, q=19, sat=False)
        self.assertAlmostEqual(p, 191.54, 2)

    def test_saturated_soil(self):
        self.soil.saturated_weight(rho_w=1000)
        self.assertEqual(self.soil.gamma_sat, 22.924)


if __name__ == '__main__':
    unittest.main()



import context
import unittest
import warnings
import numpy as np

import breakwater.core.overtopping as overtopping


class TestOvertopping(unittest.TestCase):

    def test_rubble_mound(self):
        # EurOtop (2018) case study 7 --> run 3
        with warnings.catch_warnings(record=True) as w:
            Rc = overtopping.rubble_mound(
                Hm0=2.72, q=1.18, xi_m_min_1=1.93, alpha=np.arctan(1/3),
                beta=0, gamma_b=1, gamma_v=1, gam_star=1, armour_layer='Rock',
                layers=2, permeability='permeable', safety=1, Gc=6.4,
                Dn50=1.28)

        self.assertAlmostEqual(Rc, 3.65, 2)
        self.assertEqual(len(w), 1)
        self.assertEqual(w[0].category, UserWarning)

        # EurOtop (2018) case study 8
        Rc = overtopping.rubble_mound(
            Hm0=4.9, q=21, xi_m_min_1=3.89, alpha=np.arctan(1/1.5), beta=0,
            gamma_b=1, gamma_v=1, gam_star=1, armour_layer='Accropode II',
            safety=1)
        self.assertAlmostEqual(Rc, 5.61, 2)

    def test_vertical_deep(self):
        # EurOtop (2018) case study 6, run 2
        # test of vertical_deep --> eq7.1
        with warnings.catch_warnings(record=True) as w:
            Rc = overtopping.vertical(
                Hm0=2.53, q=0.059, h=10.28, d=10.28, L_m_min_1=44,
                s_m_min_1=0.057, safety=0, limit=False)
            self.assertAlmostEqual(Rc, 5.94, 2)
            Rc = overtopping.vertical(
                Hm0=2.53, q=0.059, h=10.28, d=10.28, L_m_min_1=44,
                s_m_min_1=0.057, safety=0, limit=True)
            self.assertAlmostEqual(Rc, 5.56, 2)

        # test warnings
        self.assertEqual(len(w), 2)
        self.assertEqual(w[0].category, UserWarning)

    def test_composite_normal(self):
        # EurOtop (2018) case study 10, run 1
        # test of composite_normal --> eq7.14
        Rc = overtopping.vertical(
            Hm0=2.47, q=4.2, h=5.04, d=2.79, L_m_min_1=44,
            s_m_min_1=0.0554, safety=0)
        self.assertAlmostEqual(Rc, 5.60, 2)

    def test_vertical_no_breaking(self):
        # EurOtop (2018) case study 12, run 1
        # test of vertical_no_breaking --> eq7.5 (vertical wall)
        Rc = overtopping.vertical(
            Hm0=2.22, q=0.429, h=4.97, d=4.97, L_m_min_1=45,
            s_m_min_1=0.0497, safety=0)
        self.assertAlmostEqual(Rc, 5.67, 2)

    def test_vertical_normal(self):
        # EurOtop (2018) case study 12, run 4
        # test of vertical_normal --> eq7.8
        Rc = overtopping.vertical(
            Hm0=1.56, q=0.35, h=3.43, d=3.43, L_m_min_1=55.6,
            s_m_min_1=0.0278, safety=0)
        self.assertAlmostEqual(Rc, 7.21, 2)

    def test_vertical_low(self):
        # case study 12, run 4 adapted so that freeboard is low
        # test of vertical_low --> eq7.7
        Rc = overtopping.vertical(
            Hm0=1.56, q=16.2, h=3.43, d=3.43, L_m_min_1=55.6,
            s_m_min_1=0.0278, safety=0)
        self.assertAlmostEqual(Rc, 1.999, 3)

    def test_composite_low(self):
        # case study 10, run 1 adapted so that freeboard is low
        # test of composite_low --> eq7.15
        Rc = overtopping.vertical(
            Hm0=2.47, q=20.4, h=5.04, d=2.79, L_m_min_1=44,
            s_m_min_1=0.0554, safety=0)
        self.assertAlmostEqual(Rc, 3.298, 3)

    def test_composite_no_breaking(self):
        # case study 10, adapted so that there is no breaking
        # test of vertical_no_breaking --> eq7.5 (composite vertical)
        Rc = overtopping.vertical(
            Hm0=1.5, q=0.13, h=5.80, d=3.45, L_m_min_1=20,
            s_m_min_1=0.0532, safety=0)
        self.assertAlmostEqual(Rc, 4.156, 3)


if __name__ == '__main__':
    unittest.main()



import context
import unittest
import numpy as np

from breakwater.core.battjes import BattjesGroenendijk


class TestBattjesGroenendijk(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.battjes = BattjesGroenendijk(
            Hm0=0.13, h=0.27, slope_foreshore=np.arctan(1/100))

    def test_gammainc_upper(self):
        self.assertAlmostEqual(
            self.battjes.gammainc_upper(0.01, 0.1), 1.80324, 5)
        self.assertAlmostEqual(
            self.battjes.gammainc_upper(2, 1), 0.73575888, 8)

    def test_gamma_lower(self):
        self.assertAlmostEqual(
            self.battjes.gammainc_lower(0.01, 0.1), 97.62934376, 8)
        self.assertAlmostEqual(
            self.battjes.gammainc_lower(1.5, 1), 0.37894469, 8)

    def test_solver(self):
        self.battjes.Htr_tilde = 0.1
        H1_H2_tilde = self.battjes._solver([7.003, 1.060])
        self.assertAlmostEqual(H1_H2_tilde[0], 0.000, 3)
        self.assertAlmostEqual(H1_H2_tilde[1], 0.000, 3)

    def test_get_Hp(self):
        self.assertAlmostEqual(self.battjes.get_Hp(0.01), 0.171, 3)
        self.assertAlmostEqual(self.battjes.get_Hp(0.001), 0.192, 3)
        self.assertRaises(ValueError, self.battjes.get_Hp, P=1)

    def test_get_Hn(self):
        self.assertAlmostEqual(self.battjes.get_Hn(3), 0.135, 3)
        self.assertRaises(ValueError, self.battjes.get_Hn, N=0)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import fsolve

from ..utils.exceptions import InputError, user_warning
from ..utils.wave import shoaling_coefficient, dispersion

def goda_wave_heights(h, d, Ho, T, slope_foreshore, Ks=None, factor=1.8):
    """ Compute the design wave height

    Computes the design wave height :math:`H_{1/3}` and
    :math:`H_{max}` with the empirical formulas of Goda (2000)

    Parameters
    ----------
    h : float
        water depth [m]
    d : float
        water depth in front of the caisson, on top of the foundation [m]
    Ho : float
        deep water wave height [m]
    T : float
        wave period, Goda (2000) advises to use :math:`T_{1/3}` [s]
    slope_foreshore : tuple
        slope of the foreshore (V, H). For example a slope of 1:100 is
        defined as (1,100)
    Ks : float, optional, default: None
        non-linear shoaling coefficient. If the shoaling coefficient is
        not specified it is automatically computed with
        :py:func:`shoaling_coefficient` [-]
    factor : float, optional, default: 1.8
        Hmax is a probabilistic quantity. But to avoid possible
        confusion in design, a definite value of Hmax = 1.8 * Ks * H0
        is recommended. The value of 1.8 is, however, a recommendation
        and the user is free to choose another value, such as 1.6 or 2.0
        (Goda, 2000)
    """
    # compute the angle of the foreshore
    slope_foreshore = np.arctan(slope_foreshore[0]/slope_foreshore[1])

    # compute deep water wave length
    L_o = 9.81*T**2/(2*np.pi)

    # check if shoaling coefficient is given
    if Ks is None:
        # if not then compute the non-linear shoaling coefficient
        Ks = shoaling_coefficient(h=d, T=T, H0=Ho)
    else:
        # set the shoaling coefficient to the entered coefficient
        Ks = Ks

    if h/L_o >= 0.2:
        H13 = Ks*Ho
        Hmax = factor*Ks*Ho
    else:
        beta_0 = (0.028*(Ho/L_o)**-0.38
                       * np.exp(20*np.tan(slope_foreshore)**1.5))
        beta_1 = 0.52 * np.exp(4.2*np.tan(slope_foreshore))
        beta_max = max([0.92, (0.32*(Ho/L_o)**-0.29
                               * np.exp(2.4*np.tan(slope_foreshore)))])

        beta_0_star = (0.052*(Ho/L_o)**-0.38
                       * np.exp(20*np.tan(slope_foreshore)**1.5))
        beta_1_star = 0.63 * np.exp(3.8*np.tan(slope_foreshore))
        beta_max_star = max([1.65, (0.53*(Ho/L_o)**-0.29
                                    * np.exp(2.4*np.tan(slope_foreshore)))])

        H13 = min([beta_0*Ho + beta_1*h, beta_max*Ho, Ks*Ho])

        # water depth at a location of 5x H1/3
        hb = h + 5 * np.tan(slope_foreshore) * H13

        Hmax = min([beta_0_star*Ho + beta_1_star*hb,
                    beta_max_star*Ho,
                    factor*Ks*Ho])
    return H13, Hmax

class Goda:
    """ Compute wave pressure with the extended Goda formula (Takahasi, 2002)

    Goda (1992) analysed many of the successful and unsuccessful monolithic
    breakwaters and developed a practical formula that can be used to
    analyse the stability of monolithic breakwaters. The formula
    developed by Goda was not meant to compute the pressures for
    breaking waves (impulsive conditions), therefore Takahasi (2002)
    included an impulsive pressure coefficient in the formula.

    .. warning::
       Goda (1992) advices to avoid impulsive pressures when designing
       monolithic breakwaters.

    Parameters
    ----------
    Hs : float
        mean of the highest 1/3 of the wave heights [m].
    Hmax : float
        design wave height, equal to the mean of the highest 1/250 of
        the wave heights [m].
    h : float
        water depth [m]
    d : float
        water depth in front of the caisson, on top of the foundation [m]
    h_acc : float
        submerged depth of the caisson [m]
    hc : float
        height of the caisson above the water line [m]
    Bm : float
        width of the berm [m]
    T : float
        wave period, Goda (2000) advises to use :math:`T_{1/3}` [s]
    beta : float
        angle between direction of wave approach and a line normal to
        the breakwater [rad]
    rho : float
        density of water [kg/m³]
    slope_foreshore : float
        slope of the foreshore [rad]
    B : float, optional, default: None
        width of the monolithic breakwater [m]
    lambda_ : list, optional, default: [1, 1, 1]
        modification factors of Takahasi (2002) for alternative
        monolithic breakwater. Input must be
        \\lambda_= [:math:`\\lambda_1, \\lambda_2, \\lambda_3`].
    logger : dict, optional, default: None
        dict to log messages, must have keys 'INFO' and 'WARNINGS'

    Attributes
    ----------
    hb : float
        offshore water depth at a distance of five times Hs (=H13) [m]
    L : float
        wave length computed with the dispersion relation [m]
    eta_star : float
        the elevation to which the wave pressure is exerted [m]
    p1, p3, p4 : floats
        representative wave pressure intensities [Pa]
    pu : float
        uplift pressure [Pa]
    h_c_star : float
        elevation to which the wave pressure is exerted on the caisson,
        minimum value of hc and eta_star [m]
    B : float
        width of the monolithic breakwater [m]. None by default so it
        can be computed with :py:meth:`required_width`
    """
    def __init__(
            self, Hs, Hmax, h, d, h_acc, hc, Bm, T, beta, rho,
            slope_foreshore, B=None, lambda_=[1,1,1], logger=None):
        """ See help(Goda) for more info """
        # set dimensions as private variables
        self._h = h
        self._d = d
        self._h_acc = h_acc
        self._hc = hc
        self.rho = rho
        self.B = B

        # water depth at a location of 5x H1/3
        self.hb = h + 5 * np.tan(slope_foreshore) * Hs

        # compute the wave length with the dispersion relation
        self.L = dispersion(T=T, h=self._h)

        # compute wave pressure coefficients (Goda, 2000)
        alpha_1 = (0.6 + 0.5*((4*np.pi*self._h/self.L)
                               / np.sinh(4*np.pi*self._h/self.L))**2)
        alpha_2 = min([(self.hb-self._d)/(3*self.hb)*(Hmax/self._d)**2,
                        2*self._d/Hmax])
        alpha_3 = 1 - self._h_acc/self._h * (1-1/np.cosh(2*np.pi*self._h/self.L))

        # check for impulsive pressures and adjust alpha_2 if needed
        alpha_I = self._impulsive_pressure(Bm=Bm, Hmax=Hmax)
        if alpha_I >= alpha_2:
            alpha_star = alpha_I
            user_warning(
                ('Encountered impulsive conditions in Goda formula, it\'s '
                'recommended to change the dimensions of the structure '
                '(Goda, 2000)'))
        else:
            alpha_star = alpha_2
            if logger is not None:
                logger['INFO'].append(
                    'no impulisve conditions in Goda formula')

        # Compute the elevation to which the wave pressure is exerted
        self.eta_star = 0.75*(1 + np.cos(beta))*lambda_[0]*Hmax

        # Compute the wave pressures
        self.p1 = (0.5*(1 + np.cos(beta))
                   * (lambda_[0]*alpha_1
                      + lambda_[1]*alpha_star*np.cos(beta)**2)
                   * self.rho*9.81*Hmax)
        self.p3 = alpha_3*self.p1
        if self.eta_star > self._hc:
            self.p4 = self.p1*(1-self._hc/self.eta_star)
        else:
            self.p4 = 0
        self.pu = (0.5*(1 + np.cos(beta))
                   * lambda_[2]*alpha_1*alpha_3*self.rho*9.81*Hmax)

        # Determine h_c_star
        self.h_c_star = min([self.eta_star, self._hc])

    @property
    def _width(self):
        """ Private property of the width to check if B can be returned """
        if self.B is None:
            # B was not specified, or has not yet been computed
            # raise error
            raise InputError(
                ('B is not specified as an argument, use required_width() to '
                 'compute the required width of the caisson or specify B as '
                 'an argument'))
        else:
            # B can be returned
            return self.B

    def _impulsive_pressure(self, Bm, Hmax):
        """ Compute impulsive wave pressure coefficient

        Computes the impulsive wave pressure coefficient of the
        extended Goda formula (Takahasi, 2002)
        """
        delta_11 = (0.93*(Bm/self.L-0.12)
                    + 0.36*((self._h-self._d)/self._h-0.6))
        delta_22 = (-0.36*(Bm/self.L-0.12)
                    + 0.93*((self._h-self._d)/self._h-0.6))

        if delta_11 <= 0:
            delta_1 = 20*delta_11
        else:
            delta_1 = 15*delta_11

        if delta_22 <= 0:
            alpha_I1 = np.cos(4.9*delta_22)/np.cosh(delta_1)
        else:
            alpha_I1 = 1/(np.cosh(delta_1)*np.sqrt(np.cosh(3*delta_22)))

        if Hmax <= 2*self._d:
            alpha_I0 = Hmax/self._d
        else:
            alpha_I0 = 2

        alpha_I = alpha_I0*alpha_I1
        return alpha_I

    def _pressure_centroid(self):
        """ Compute the centroid of the pressure

        Reference of the centroid is the lower right of the pressure
        """
        # compute total area
        A_tot = self.p1 * (self._hc + self._h_acc)

        # compute areas to subtract from total area
        A1 = 0.5 * (self.p1 - self.p4) * self._hc
        A2 = 0.5 * (self.p1 - self.p3) * self._h_acc

        # compute centroid of each area, reference is lower right
        y_C_tot = 0.5 * (self._hc + self._h_acc)
        y_C1 = self._h_acc + 2/3 * self._hc
        y_C2 = 1/3 * self._h_acc

        # compute centroid of pressures
        y = (A_tot*y_C_tot - A1*y_C1 - A2*y_C2)/(A_tot - A1 - A2)

        return y

    def _dFv(self, M):
        """ Compute net vertical force """
        # compute net vertical force, down is positive
        return M*9.81 - self.U()

    def effective_width(self, M):
        """ Compute the effective width

        The effective width must be used for geotechnical computations,
        due to the fact that the net vertical force of the caisson is
        eccentric.

        Parameters
        ----------
        M : float
            mass of the caisson [kg]

        Returns
        -------
        float
            the effective width [m]
        """
        return self._width - 2 * self.eccentricity(M)

    def eccentricity(self, M):
        """ Compute the eccentricity of the net vertical force

        Parameters
        ----------
        M : float
            mass of the caisson [kg]

        Returns
        -------
        float
            eccentricity of the net vertical force [m]
        """
        # compute and return eccentricity of the net vertical force
        return self.Ma()/self._dFv(M)

    def P(self):
        """ Compute horizontal force due to the pressure

        Returns
        -------
        float
            horizontal force due to the pressures [Pa]
        """
        P = (0.5*(self.p1+self.p3)*self._h_acc
             + 0.5*(self.p1+self.p4)*self.h_c_star)
        return P

    def Mp(self):
        """ Compute moment at the heel due to the pressure

        Returns
        -------
        float
            moment around the heel due to the horizontal pressures [Nm]
        """
        Mp = (1/6 * (2*self.p1 + self.p3)*self._h_acc**2
              + 0.5*(self.p1 + self.p4)*self._h_acc*self.h_c_star
              + 1/6 * (self.p1 + 2*self.p4)*self.h_c_star**2)
        return Mp

    def U(self):
        """ Compute force due to the uplift pressure

        Returns
        -------
        float
            vertical uplift pressure [Pa]
        """
        return 0.5*self.pu*self._width

    def Mu(self):
        """ Compute moment at the heel due to the uplift

        Returns
        -------
        float
            moment around the heel due to the uplift pressure [Nm]
        """
        return 2/3* self.U() * self._width

    def Ma(self):
        """ Compute the moment around the center of the caisson

        .. warning::
           This method assumes a symmetric caisson

        Returns
        -------
        float
            moment around the center of the caisson [Nm]
        """
        # compute and return moment, clockwise is positive
        return self.U() * self._width/6 + self.P() * self._pressure_centroid()

    def required_mass(
            self, mu, t=0.5, SF_sliding=1.2, SF_turning=1.2, logger=None):
        """ Compute required mass of the monolithic breakwater

        Compute the minimal required mass of the monolithic breakwater
        based on the failure mechanisms sliding and overturning.

        .. math::
           M_{sliding} = \\frac{P SF_{sliding}}{g \\mu} + \\frac{U}{g}
           + \\rho B h'

        .. math::
           M_{turning} = \\frac{M_p SF_{turning}}{g t} + \\frac{M_u}{g t}
           + \\rho B h'

        Parameters
        ----------
        mu : float
            friction factor between the caisson and the foundation [-]
        t : scalar, optional, default: 0.5
            horizontal distance to the centre of gravity [m]
        SF_sliding : float, optional, default: 1.2
            safety factor against sliding. Default value according to
            Goda (2000)
        SF_turning : float, optional, default: 1.2
            safety factor against sliding. Default value according to
            Goda (2000)
        logger : dict, optional, default: None
            dict to log messages, must have keys 'INFO' and 'WARNINGS'

        Returns
        -------
        mass : float
            minimal required mass per meter length to satisfy the
            safety factors [kg/m]
        """
        t = t*self._width

        # Compute forces and moments
        P = self.P()
        Mp = self.Mp()
        U = self.U()
        Mu = self.Mu()

        # Compute the required mass for sliding and turning
        mass_sliding = (P*SF_sliding/(9.81*mu)
                        + U/9.81 + self.rho*self._width*self._h_acc)
        mass_turning = (Mp*SF_turning/(9.81*t)
                        + Mu/(t*9.81) + self.rho*self._width*self._h_acc)

        # normative mass is largest mass of the two
        if mass_sliding >= mass_turning:
            mass_required = mass_sliding
            if logger is not None:
                logger['INFO'].append(
                    ('safety factor for sliding is normative for the '
                     'computation of the mass'))
        else:
            mass_required = mass_turning
            if logger is not None:
                logger['INFO'].append(
                    ('safety factor for overtuning is normative for the '
                     'computation of the mass'))

        return mass_required

    def required_width(
            self, Pc, rho_c, rho_f, rho_w, mu, t=0.5,
            SF_sliding=1.2, SF_turning=1.2, logger=None):
        """ Compute the required width of the monolithic breakwater

        Compute the minimal required width of the monolithic breakwater
        based on the failure mechanisms sliding and overturning.

        Parameters
        ----------
        Pc : float
            contribution of concrete to the total mass of the caisson.
            value between 0 and 1
        rho_c : float
            density of concrete [kg/m³]
        rho_f : float
            density of the fill material, for instance sand [kg/m³]
        rho_w : float
            density of water [kg/m³]
        mu : float
            friction factor between the caisson and the foundation [-]
        t : scalar, optional, default: 0.5
            horizontal distance to the centre of gravity [m]
        SF_sliding : float, optional, default: 1.2
            safety factor against sliding. Default value according to
            Goda (2000)
        SF_turning : float, optional, default: 1.2
            safety factor against sliding. Default value according to
            Goda (2000)
        logger : dict, optional, default: None
            dict to log messages, must have keys 'INFO' and 'WARNINGS'

        Returns
        -------
        B : float
            minimal required width to satisfy the safety factors [m]
        """
        # Compute horizontal force and moment due to this force
        P = self.P()
        Mp = self.Mp()

        # contribution of the fill material to the total mass
        Pf = 1 - Pc

        # compute the mass per meter length and width
        m_acc = (self._hc * (Pf*rho_f + Pc*rho_c)
                 + self._h_acc * (Pf*(rho_f-rho_w) + Pc*(rho_c-rho_w)))

        # compute the minimal required width for sliding and overturning
        B_sliding = 2*SF_sliding*P/(mu*(2*m_acc*9.81 - self.pu))
        B_turning = np.sqrt(SF_turning*Mp
                            / (0.5*m_acc*9.81 - (2*t/3)*self.pu))

        # normative value is largest required width
        if B_sliding >= B_turning:
            B_required = B_sliding
            if logger is not None:
                logger['INFO'].append(
                    ('safety factor for sliding is normative for the '
                     'computation of the width'))
        else:
            B_required = B_turning
            if logger is not None:
                logger['INFO'].append(
                    ('safety factor for overtuning is normative for the '
                     'computation of the width'))

        # update attribute, first check if one wass specified
        if self.B is not None:
            # B was not None, show warning
            user_warning(
                f'B was given as {self.B}, now changed to {B_required}')

        # update B
        self.B = B_required

        return B_required

    def bearing_pressure_width(self, B1, Pc, rho_c, rho_fill, pe_max, t=0.5):
        """ Compute the required width for the bearing pressure

        Method uses fsolve from scipy to compute the width that
        satisfy the maximum bearing pressure of the foundation.

        Parameters
        ----------
        B1 : float
            first estimate for the width [m]
        Pc : float
            contribution of concrete to the total mass of the caisson.
            value between 0 and 1
        rho_c : float
            density of concrete [kg/m³]
        rho_f : float
            density of the fill material, for instance sand [kg/m³]
        pe_max : float
            maximum value of the bearing pressure at the heel of the
            caisson. Goda (2000) advises a value between 400 and 500
            kPa.
        t : scalar, optional, default: 0.5
            horizontal distance to the centre of gravity [m]

        Returns
        -------
        float
            required width for the maximum bearing pressure [m]
        """
        # define lambda function for fsolve and compute new width
        func = lambda B: (self.bearing_pressure(
                B=B, Pc=Pc, rho_c=rho_c, rho_fill=rho_fill)
            - pe_max*1000)
        B = fsolve(func, B1, xtol=0.02)[0]

        # update attribute
        self.B = B

        return B

    def bearing_pressure(self, Pc, rho_c, rho_fill, t=0.5, B=None):
        """ compute the bearing pressure at the heel

        Method to compute the bearing pressure at the heel of the
        caisson.

        .. math::
           p_{e}=\\left\\{\\begin{array}{ll} \\frac{2 W_{e}}{3 t_{e}} & :
           t_{e} \\leq \\frac{1}{3} B \\\ \\frac{2 W_{e}}{B}\\left(2-3
           \\frac{t_{e}}{B}\\right) & : t_{e}>\\frac{1}{3} B
           \\end{array}\\right.

        in which:

        .. math::
           t_{e}=\\frac{M_{e}}{W_{e}}, \\quad M_{e}=M g t-M_{U}-M_{p},
           \\quad W_{e}=M g-U

        Parameters
        -------
        Pc : float
            contribution of concrete to the total mass of the caisson.
            value between 0 and 1
        rho_c : float
            density of concrete [kg/m³]
        rho_f : float
            density of the fill material, for instance sand [kg/m³]
        t : scalar, optional, default: 0.5
            horizontal distance to the centre of gravity [m]
        B : float, optional, default: None
            width of the monolithic breakwater [m], used with
            :py:meth:`bearing_pressure_width` to compute the required
            width to satisfy the maximum bearing pressure.

        Returns
        -------
        pe : float
            the bearing pressure at the heel of the caisson [Pa]
        """
        if B is None:
            # get the width
            B = self._width

        # compute the mass of the structure
        M = self.mass(Pc=Pc, rho_c=rho_c, rho_fill=rho_fill)

        W = (M - self.rho*B*self._h_acc) * 9.81

        We = W - self.U()
        Me = W*t*B - self.Mu() - self.Mp()
        te = Me/We

        # compute bearing pressure
        if te <= B/3:
            pe = 2*We/(3*te)
        else:
            pe = 2*We/B * (2 - 3*te/B)

        return pe

    def mass(self, Pc, rho_c, rho_fill):
        """ Compute the mass of the caisson

        Parameters
        ----------
        Pc : float
            contribution of concrete to the total mass of the caisson.
            value between 0 and 1
        rho_c : float
            density of concrete [kg/m³]
        rho_f : float
            density of the fill material, for instance sand [kg/m³]

        Returns
        -------
        m : float
            minimal required mass per meter length to satisfy the
            safety factors [kg/m]
        """
        # compute area
        A = self._width * (self._h_acc + self._hc)

        # compute the mass
        M = A*Pc*rho_c + A*(1-Pc)*rho_fill

        return M

    def plot(self):
        """ Plot pressure distribution

        Plots the pressure distribution together with the dimensions of
        the monolithic breakwater.

        .. warning::
           Do not read the dimensions of the monolithic breakwater from
           the axes of the figure. The correct dimensions of the
           monolithic breakwater can be read from the figure.
        """
        # get the width
        B = self._width

        p = np.array([self.p3,
                      self.p1,
                      self.p1-self.p1*self._hc/self.eta_star])/1000
        y = np.array([0, self._h_acc, self._h_acc+self._hc])
        x = np.array([0, B])
        pu = np.array([self.pu, 0])/1000

        # scale the size of the caisson with the pressure
        scale = np.max(p)/np.max(y)
        y = y*scale
        x = x*scale

        # plot the caisson and wlev
        plt.vlines(x=0, ymin=0, ymax=-y[2], linewidth=2)
        plt.vlines(x=-x[1], ymin=0, ymax=-y[2], linewidth=2)
        plt.hlines(y=0, xmin=0, xmax=-x[1], linewidth=2)
        plt.hlines(y=-y[2], xmin=0, xmax=-x[1], linewidth=2)

        plt.hlines(y=-self._h_acc*scale, xmin=np.max(p)*1.3, xmax=0, color='b')

        # plot dimensions of the caisson
        plt.text(
            x=(-x[1]/2)*1.1, y=(-y[2]/2)*1.03, s=f'B = {np.round(B, 1)} m')
        plt.arrow(x=-x[1]/2, y=-y[2]/2, dx=x[1]/2, dy=0, head_width=4,
                  head_length=4, color='gray', length_includes_head=True)
        plt.arrow(x=-x[1]/2, y=-y[2]/2, dx=-x[1]/2, dy=0, head_width=4,
                  head_length=4, color='gray', length_includes_head=True)

        plt.text(x=(-x[1]/2)*0.9, y=(-y[2]/2)*1.1,
                 s=f'L = {np.round(y[2]/scale, 2)} m', rotation=90)
        plt.arrow(x=-x[1]/2, y=-y[2]/2, dx=0, dy=-y[2]/2, head_width=4,
                  head_length=4, color='gray', length_includes_head=True)
        plt.arrow(x=-x[1]/2, y=-y[2]/2, dx=0, dy=y[2]/2, head_width=4,
                  head_length=4, color='gray', length_includes_head=True)

        # plot pressure distributions
        plt.hlines(y=0, xmin=0, xmax=p[0], color='deepskyblue')
        plt.hlines(y=-y[2], xmin=0, xmax=p[2], color='deepskyblue')
        plt.plot(p, -y, color='deepskyblue')

        plt.vlines(x=0, ymin=0, ymax=pu[0], color='deepskyblue')
        plt.plot(-x, pu, color='deepskyblue')

        # hide negative ticks
        xticks = [xtick for xtick in plt.gca().get_xticks() if xtick >= 0]
        yticks = [ytick for ytick in plt.gca().get_yticks() if ytick >= 0]

        plt.gca().set_xticks(xticks)
        plt.gca().set_yticks(yticks)

        # invert axes
        plt.xlim(np.max(p)*1.3, -np.max(x)-5)
        plt.ylim(np.max(pu)+5, -np.max(y)-5)

        # add title and label
        plt.title('Pressure distributions computed with Goda (2000)')
        plt.xlabel('Pressure [kPa]')
        plt.ylabel('Pressure [kPa]')

        # add grid and show plot
        plt.grid()
        plt.show()



import numpy as np
import matplotlib.pyplot as plt

from .exceptions import InputError, RockGradingError

def _process_cost(structure, cost, Grading, validate=True):
    """ Process cost input to a dict

    Parameters
    ----------
    structure : {'RRM', 'CRM', 'RC', 'CC'}
        structure for which the cost must be verified
    Grading : :py:class:`RockGrading`
        rock grading
    cost : dict
        dictionary with the cost
    validate : bool, optional, default: True
        if True the input is validated, if False the cost dict is
        returned
    """
    # check if cost have been specified
    if cost is not None:
        # cost have been added
        # check if cost have been added to the grading
        if 'price' in Grading[list(Grading.grading.keys())[0]]:
            # pricing has been added
            pass
        else:
            # pricing has not been added, raise error
            raise RockGradingError('There is no pricing in the RockGrading')

        # check structure to check structure specific cost
        if 'RRM' in structure:
            # check if core_price is in cost
            if 'core_price' not in cost.keys() or cost['core_price'] is None:
                raise KeyError(
                    'core_price must be specified when computing the cost of RRM')

        if 'CRM' in structure or 'CC' in structure:
            # check if unit_price is in cost
            if 'unit_price' not in cost.keys() or cost['unit_price'] is None:
                raise KeyError(
                    'unit_price must be specified when computing the cost of CRM/CC')

        if 'RC' in structure or 'CC' in structure:
            # check if fill_price is in cost
            if 'fill_price' not in cost.keys() or cost['fill_price'] is None:
                raise KeyError(
                    'fill_price must be specified when computing the cost of CRM/CC')

            # check if concrete_price is in cost
            if 'concrete_price' not in cost.keys() or cost['concrete_price'] is None:
                raise KeyError(
                    'concrete_price must be specified when computing the cost of CRM/CC')


        # check if cost must be returned
        if not validate:
            # check for optional parameters
            if 'transport_price' not in cost:
                # add as None
                cost['transport_price'] = None

            if 'dry_dock' not in cost:
                # add as None
                cost['transport_cost'] = None

            if 'length' not in cost:
                cost['length'] = None

            return cost

def cost_influence(lines):
    """ Plot influence of varying parameters

    Parameters
    ----------
    lines : dict
        dictionary with the parameters as keys and a nested dict with
        the values and cost
    """
    # check if more than one parameter has been given
    if len(lines.keys()) > 1:
        # values must be normalised
        normalise = True

        # set title
        title = 'Influence of the varying parameters on the cost (normalised)'
        xlabel = 'normalised value of the parameter, from min to max'

        # set xlim
        xmin, xmax = 0, 1

    else:
        # normalising is not needed
        normalise = False

        # set title
        title = 'Influence of the varying parameters on the cost'

    # create the figure
    plt.figure()

    # iterate over the lines
    for parameter, data in lines.items():
        # set empty list for x data
        x = []

        # check if data must be normalised
        if normalise:
            # get min and max value
            min = np.min(data['values'])
            max = np.max(data['values'])

            for value in data['values']:
                # normalise data and append to list
                x.append((value - min)/(max - min))

            # add min and max to label
            label = f'{parameter} (min={min}, max={max})'

        else:
            # x equals the values
            x = data['values']

            # label is parameter
            label = parameter
            xlabel = parameter

            # set xmax and xmax
            xmin, xmax = np.min(x), np.max(x)

        # plot data
        plt.plot(x, data['cost'], label=label)

    # style figure
    plt.xlim(xmin, xmax)
    plt.xlabel(xlabel)
    plt.ylabel('Cost per m')
    plt.title(title)
    plt.legend()
    plt.grid()
    plt.tight_layout()
    plt.show()



import numpy as np
from scipy.optimize import fsolve

def shoaling_coefficient(h, T, H0, linear=False):
    """ Closed form equation for the non-linear shoaling coefficient

    Computes the non-linear shoaling coefficient using a closed form
    equation. The equation uses the coefficient derived in Kweon and
    Goda (1996).

    Parameters
    ----------
    h : float
        water depth [m]
    T : float
        wave period, Goda (2000) advises to use :math:`T_{1/3}` [s]
    H0 : float
        equivalent deep water significant wave height [m]
    linear : bool, optional, default: False
        if true the linear shoaling coefficient is returned, if false
        the non-linear shoaling coefficient is returned

    Returns
    -------
    Ks : float
        the non-linear shoaling coefficient
    """
    # set gravitational acceleration
    g = 9.81

    # compute wave length with the dispersion relation
    L = dispersion(T=T, h=h)
    L0  = g*T**2/(2*np.pi)

    # compute wave number
    k = 2*np.pi/L

    # compute ratio c_g (group velocity) to c (wave celerity)
    n = 0.5*(1 + 2*k*h/np.sinh(2*k*h))

    # compute the linear shoaling coefficient
    Ksi = np.sqrt(1/np.tanh(k*h) * 1/(2*n))

    if linear:
        return Ksi
    else:
        # compute the non-linear shoaling coefficient
        K = Ksi + 0.0015 * (h/L0)**-2.87 * (H0/L0)**1.27
        return K


def dispersion(T, h):
    """ Dispersion relationship

    uses fsolve to find the wave length

    Parameters
    ----------
    T : float
        wave period [s]
    h : float
        water depth [m]

    Returns
    L : float
        the wave length [m]
    """
    g = 9.81

    # deep water wave length
    L0  = g*T**2/(2*np.pi)

    # define dispersion relation with a lambda function to solve L
    dispersion = lambda L : L0 * np.tanh(2*np.pi*h/L) - L

    # solve for L
    L = fsolve(dispersion, L0)[0]

    return L



import context
import unittest
import warnings
import numpy as np

from breakwater.conditions import LimitState
from breakwater.utils.exceptions import LimitStateWarning


class TestLimitState(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ULS = LimitState(
            h=8.5, Nod=0.5, q=10, Hm0=3.22, H13=3.27, T_m_min_1=7.66, Tp=6.96,
            Tm=6.33, label='ULS')
        cls.SLS = LimitState(
            h=8.5, Sd=10, Hm0=2.82, T_m_min_1=8.58, label='SLS')

    def test_deep_water(self):
        self.assertFalse(self.ULS.deep_water)
        self.assertFalse(self.SLS.deep_water)

    def test_transform_periods(self):
        with warnings.catch_warnings(record=True) as w:
            self.SLS.transform_periods(scalar=0.5)

        self.assertEqual(len(w), 1)
        self.assertEqual(w[0].category, LimitStateWarning)

        parameter = ['Sd', 'Hm0', 'T_m_min_1', 'Tp', 'Tm', 'T13']
        computed = [10, 2.82, 8.58, 7.8, 6.24, 7.215]
        for i, (arg, val) in enumerate(self.SLS.conditions.items()):
            self.assertEqual(arg, parameter[i])
            self.assertEqual(val, computed[i])

    def test_get_Hs(self):
        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(self.ULS.get_Hs('H13'), 3.27)
            self.assertEqual(self.ULS.get_Hs('Hm0'), 3.22)
            self.assertEqual(self.SLS.get_Hs('H13'), 2.82)

        self.assertEqual(len(w), 1)
        self.assertEqual(w[0].category, LimitStateWarning)

    def test_L(self):
        L0 = self.ULS.L(period='T_m_min_1', deep_water=True)
        L = self.ULS.L(period='T_m_min_1', deep_water=False)
        self.assertAlmostEqual(L0, 91.611, 3)
        self.assertAlmostEqual(L, 63.122, 3)

    def test_s(self):
        self.assertAlmostEqual(self.ULS.s(number='mean'), 0.05227, 5)
        self.assertAlmostEqual(self.ULS.s(number='spectral'), 0.03515, 5)
        self.assertAlmostEqual(self.ULS.s(H='Hm0', T='T_m_min_1'), 0.03515, 5)

    def test_surf_similarity(self):
        alpha = np.arctan(3/4)
        xi_mean = self.ULS.surf_similarity(alpha=alpha, number='mean')
        xi = self.ULS.surf_similarity(alpha=alpha, H='Hm0', T='T_m_min_1')
        self.assertAlmostEqual(xi_mean, 3.280, 3)
        self.assertAlmostEqual(xi, 4.000, 3)

    def test_Sd(self):
        self.ULS.Sd(G=1, nv=0.45)
        self.assertAlmostEqual(self.ULS.conditions['Sd'], 0.909, 3)

    def test_Nod(self):
        LS = LimitState(
            h=8.5, Sd=10, Hm0=2.82, T_m_min_1=8.58, label='LS')

        LS.Nod(G=1, nv=0.45)
        self.assertEqual(LS.conditions['Nod'], 5.5)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
from tabulate import tabulate
from pandas import read_excel, read_csv

from .utils.exceptions import InputError, RockGradingError, ArmourUnitsError, user_warning


def read_grading(filepath, rho=2650, sheet_name=0):
    """ Load a rock grading from an excel or csv file

    Excel input file can be generated by using
    :obj:`bw.generate_excel <breakwater.utils.input_generator.generate_excel>`
    with input argument 'Grading'. Alternatively, when using your own
    Excel file it must, at least, have the following headers.

    +------------+-----------+-----------+-----+-----+
    | Rock Class | M50 Lower | M50 Upper | NLL | NUL |
    +------------+-----------+-----------+-----+-----+

    Parameters
    ----------
    filepath : str
        a valid filepath
    rho : float, optional, default: 2650
        density of the armourstone [kg/m³]
    sheet_name : str, optional, default: 0
        name of the sheet with the RockGrading to read, default value
        of 0 will result in reading the first sheet

    Returns
    -------
    bw.RockGrading
        a rock grading object
    """
    extension = filepath.split('.')[-1]

    if extension == 'xlsx':
        df = read_excel(filepath, sheet_name=sheet_name)

    elif extension == 'csv':
        df = read_csv(filepath, delimiter=";|,", engine='python')

    else:
        raise IOError(f'specified file has an invalid extension')

    # create custom_grading
    custom_grading = {}

    # lowercase all columns
    df.columns = map(str.lower, df.columns)

    # iterate over the rows in the df
    for i, row in df.iterrows():
        rock_class = row['Rock Class'.lower()]
        custom_grading[rock_class] = {

            'M50': [row['M50 Lower'.lower()], row['M50 Upper'.lower()]],

            'NLL': row['NLL'.lower()],

            'NUL': row['NUL'.lower()]
        }

    return RockGrading(grading=custom_grading, rho=rho)

def read_units(filepath, kd, name, rho=2400, sheet_name=0):
    """ Load ArmourUnits from an excel or csv file

    Excel input file can be generated by using
    :obj:`bw.generate_excel <breakwater.utils.input_generator.generate_excel>`
    with input argument 'ArmourUnits'. Alternatively, when using your
    own Excel file it must, at least, have the following headers.

    +-----------+-------+-------+--------+
    | Volume    |   D   |   h   |   Vc   |
    +-----------+-------+-------+--------+

    .. note::
       Do not use this function to read Xbloc or XblocPlus armour units
       from an Excel file. Use the predefined,
       :obj:`bw.Xbloc <breakwater.material.Xbloc>` or
       :obj:`bw.XblocPlus <breakwater.material.XblocPlus>` instead.


    Parameters
    ----------
    filepath : str
        a valid filepath
    kd : int
        Stability coefficient [-]
    name : str
        name of the ArmourUnit
    rho : float, optional, default: 2400
        density of the concrete [kg/m³]
    sheet_name : str
        name of the sheet with the armour units to read, default value
        of 0 will result in reading the first sheet

    Returns
    -------
    bw.ConcreteArmour
        an armour units object
    """
    extension = filepath.split('.')[-1]

    if extension == 'xlsx':
        df = read_excel(filepath, sheet_name=sheet_name)

    elif extension == 'csv':
        df = read_csv(filepath, delimiter=";|,", engine='python')

    else:
        raise IOError(f'specified file has an invalid extension')


    if name in ['Xbloc', 'XblocPlus']:
        user_warning(
            (f'The correction factor for {name} is not computed when '
             f'loading {name} from an Excel file.'))

    # create custom_grading
    units = {}

    # lowercase all columns
    df.columns = map(str.lower, df.columns)

    for i, row in df.iterrows():
        class_ = row['Volume'.lower()]
        units[class_] = {

            'D': row['D'.lower()],

            'h': row['h'.lower()],

            'Vc': row['Vc'.lower()]
        }

    return ConcreteArmour(kd=kd, units=units, name=name, rho=rho)

class RockGrading:
    """ Standard rock grading of the standard NEN-EN 13383-1 (2002)

    Create a Rock Grading to determine the rock class in
    :py:class:`RockRubbleMound`, :py:class:`ConcreteRubbleMound`,
    :py:class:`RubbleMound` or :py:class:`Caisson`. By default the
    standard rock grading of the NEN-EN 13383-1 (2002) is used.
    However, it is possible to change the grading to a user defined
    custom grading by giving one in the arguments.

    Parameters
    ----------
    grading : dict, optional, default: None
        If specified the custom grading will replace the default rock
        grading of the NEN-EN 13383-1. The custom grading must be
        formatted as {str: {'M50': [float, float], 'NLL': float, 'NUL':
        float}}, with str the rock class and a lower and upper bound
        for M50. Furthermore, the NLL and NUL mass must be specified.
    rho : float, optional, default: 2650
        density of the armourstone [kg/m³]
    y_NLL : float, optional, default: 0.06
        the fraction passing at the nominal lower limit mass (NLL) of a
        grading
    y_NUL : float, optional, default: 0.9
        the fraction passing at the nominal upper limit mass (NUL) of a
        grading

    Attributes
    ----------
    grading : dict
        dictionary of the chosen rock grading
    rho : float
        density of the armourstone
    y_NLL : float
        the fraction passing at the nominal lower limit mass (NLL)
    y_NUL : float
        the fraction passing at the nominal upper limit mass (NUL)
    """
    def __init__(self, grading=None, rho=2650, y_NLL=0.06, y_NUL=0.9):
        """ See help(RockGrading) for more info """
        self.name = 'Rock'
        self.rho = rho
        self.y_NLL = y_NLL
        self.y_NUL = y_NUL

        if grading is not None:
            # check if format of rock grading is a valid format
            msg = ('Invalid custom rock grading. Rock grading must be '
                   'formatted as {str: {\'M50\': [float, float], \'NLL\': '
                   'float, \'NUL\': float}}')
            if isinstance(grading, dict):
                # grading is a dict
                for class_, values in grading.items():
                    if isinstance(values, dict):
                        # grading is a nested dict
                        for i, (param, value) in enumerate(values.items()):
                            if isinstance(value, list) and len(value) == 2:
                                # mass is list with an upper and lower bound
                                if param == 'M50':
                                    # key is M50
                                    pass
                                else:
                                    # invalid key for M50
                                    raise RockGradingError(msg)
                            elif param == 'NLL':
                                pass
                            elif param == 'NUL':
                                pass
                            else:
                                raise RockGradingError(msg)
                    else:
                        # grading is not a nested dict
                        raise RockGradingError(msg)
            else:
                # grading is not a dict
                raise RockGradingError(msg)
        else:
            # no custom grading was specified, use NEN-EN 13383-1
            grading = {'LMA_5/40':
                           {'M50': [10, 20], 'NLL': 5, 'NUL': 40},
                       'LMA_10/60':
                           {'M50': [20, 35], 'NLL': 10, 'NUL': 60},
                       'LMA_40/200':
                           {'M50': [80, 120], 'NLL': 40,'NUL': 200},
                       'LMA_15/300':
                           {'M50': [45, 135], 'NLL': 15, 'NUL': 300},
                       'LMA_60/300':
                           {'M50': [120, 190], 'NLL': 60, 'NUL': 300},
                       'HMA_300/1000':
                           {'M50': [540, 690], 'NLL': 300, 'NUL': 1000},
                       'HMA_1000/3000':
                           {'M50': [1700, 2100], 'NLL': 1000, 'NUL': 3000},
                       'HMA_3000/6000':
                           {'M50': [4200, 4800], 'NLL': 3000, 'NUL': 6000},
                       'HMA_6000/10000':
                           {'M50': [7500, 8500], 'NLL': 6000, 'NUL': 10000},
                       'HMA_10000/15000':
                           {'M50': [12000, 13000], 'NLL': 10000, 'NUL': 15000}}

        # set the grading as an attribute
        self.grading = grading

    def __str__(self):
        table = []

        for i, (class_, values) in enumerate(self.grading.items()):
            if i == 0:
                # set table headers in first iteration
                headers = ['Rock Class']
                # get the keys of the nested dict and add to headers
                headers.extend(list(values.keys()))

            # add the rock class to a new row
            row = [class_]

            # add the values to the row
            row.extend(list(values.values()))

            # add row to the table
            table.append(row)

        # return the table as a string for printing
        return str(tabulate(table, headers, tablefmt='github'))

    def __setitem__(self, key, value):
        self.grading[key] = value

    def __getitem__(self, key):
        return self.grading[key]

    def add_cost(self, cost):
        """ Add cost per m³ of each Rock Class

        Add the cost per m³ of each Rock Class to the :py:attr:`grading`

        Parameters
        ----------
        cost : dict
            cost of each Rock Class, keys must be identical to the name
            of the rock class

        Raises
        ------
        KeyError
            if a rock class of the given cost is not in the grading
        InputError
            if a rock class of the :py:attr:`grading` has no price
        """
        rock_classes = list(self.grading.keys())
        # iterate over the prices in the given dict
        for class_, price in cost.items():
            # check if the rock class is in the grading
            if class_ in self.grading.keys():
                # class is in the grading so add price to the nested dict
                self.grading[class_]['price'] = price

                # delete from rock_classes to validate if all have a price
                rock_classes.remove(class_)
            else:
                # class not in the grading, raise KeyError
                valid_classes = ', '.join(self.grading.keys())
                raise KeyError(
                    (f'{class_} is not a rock class of the RockGrading, '
                     f'valid classes are {valid_classes}'))

        # check if pricing has been added for all rock_classes
        if any(rock_classes):
            # not all have been added, raise error
            no_pricing = ', '.join(rock_classes)
            raise InputError(f'No pricing has been given for {no_pricing}')

    def get_class(self, Dn50):
        """ Get the rock class for a given Dn50

        Parameters
        ----------
        Dn50 : float
            nominal diameter of the armourstone [m]

        Returns
        -------
        str
            Rock class

        Raises
        ------
        RockGradingError
            If the computed Dn50 of the armour layer is out of range for
            the specified rock grading
        """
        rock_class = None
        M_required = self.rho * Dn50**3

        for class_, characteristics in self.grading.items():
            if characteristics['M50'][1] >= M_required:
                rock_class = class_
                break

        if rock_class == None:
            max_class = list(self.grading.keys())[-1]
            max_mass = self.grading[max_class]['M50'][1]
            max_dn = np.round((max_mass/2650)**(1/3), 3)
            raise RockGradingError(
                f'Dn50 = {np.round(Dn50, 3)} is out of range for the specified'
                f' rock grading, {max_dn} m is the maximum possible Dn50.')
        else:
            return rock_class

    def get_class_dn50(self, class_):
        """ Get the average Dn50 of a rock class

        Parameters
        ----------
        class_ : str
            Rock class

        Returns
        -------
        Dn50 : float
            nominal diameter of the armourstone [m]

        Raises
        ------
        KeyError
            If the specified rock class is not in the rock grading
        """
        class_M = self.grading[class_]['M50']

        average_Dn50 = ((class_M[0] + class_M[1])/(2*self.rho))**(1/3)

        return average_Dn50

    def rosin_rammler(self, class_, y):
        """ Compute the Rosin-Rammler curve for an idealised gradings

        The Rosin-Rammler curve can be used to interpolate between the
        limits of standard gradings (CIRIA, CUR, CETMEF, 2007). Given
        that two fixed point on the curve are known, for instance the
        NLL and NUL limits, M50 and  the uniformity coefficient, n_RRM,
        can be computed. These values are subsequently used to compute
        My.

        .. math::
           M_{y} = M_{50} \\left(\\frac{\\ln{(1 - y)}}{\\ln{(0.5)}}
           \\right)^{1/n_{RRM}}

        Parameters
        ----------
        class_ : str
            Rock class
        y : float
            the fraction passing value

        Returns
        -------
        My : float
            the mass corresponding to that value using a percentage
            subscript to express that fraction [kg]

        Raises
        ------
        KeyError
            If the specified rock class is not in the rock grading
        """
        # set NLL and NUL for the given rock class
        NLL = self.grading[class_]['NLL']
        NUL = self.grading[class_]['NUL']

        # compute the uniformity index
        a = np.log(1 - self.y_NUL)/np.log(1 - self.y_NLL)
        n_RRM = np.log10(a)/np.log10(NUL/NLL)

        # compute M50 with NLL and the computed n_RRM, in accordance
        # with the advice from the Rock Manual (p113, bottom)
        M50 = NLL * (np.log(1 - self.y_NLL)/np.log(0.5))**(-1/n_RRM)

        # compute the mass passing
        My = M50 * (np.log(1- y)/np.log(0.5))**(1/n_RRM)

        return My

    def plot_rosin_rammler(self, class_):
        """ plot the Rosin-Rammler curve for an idealised gradings

        Parameters
        ----------
        class_ : str
            Rock class

        Raises
        ------
        KeyError
            If the specified rock class is not in the rock grading
        """
        y = np.linspace(0, 0.999999, 1000)
        My = self.rosin_rammler(class_=class_, y=y)

        # plot grading curve and NLL + NUL points
        plt.plot(My, y)
        plt.plot(self.grading[class_]['NLL'], self.y_NLL, 'ko')
        plt.plot(self.grading[class_]['NUL'], self.y_NUL, 'ko')

        # style plot
        xmax = self.rosin_rammler(class_=class_, y=0.9999)
        xmin = self.rosin_rammler(class_=class_, y=0.0001)
        plt.xlim(xmin=xmin, xmax=xmax)
        plt.ylim(ymin=0, ymax=1)

        plt.xlabel('My, Mass of armourstone [kg]')
        plt.ylabel('y, fraction lighter')

        plt.title(f'Rosin-Rammler curve for {class_}')

        plt.grid('minor')
        plt.show()


class ConcreteArmour:
    """ Define concrete armour units

    Define a type of concrete armour unit, :py:class:`Xbloc` and
    :py:class:`XblocPlus` have been predefined.

    Parameters
    ----------
    kd : int
        Stability coefficient [-]
    units : dict
        Dictionary of the defined concrete armour units. Format must
        be {V: dict}, where V is the volume of the armour unit (float)
        and dict = {'D': float, 'h': float, 'Vc': float} in which D is
        the diameter, h is the thickness of the armour layers and Vc is
        the volume of concrete.
    name : str, optional, default: None
        name of the ArmourUnit, by default the name of the armour unit
        is derived from the name of the class.
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units [kg/m³]

    Attributes
    ----------
    kd : int
        Stability coefficient [-]
    units : dict
        Dictionary of the defined concrete armour units. Format must
        be {V: dict}, where V is the volume of the armour unit (float)
        and dict = {'D': float, 'h': float, 'Vc': float} in which D is
        the diameter, h is the thickness of the armour layers and Vc is
        the volume of concrete.
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units [kg/m³]
    """

    def __init__(self, kd, units, name=None, rho=2400):
        """ See help(ConcreteArmour) for more info """
        self.kd = kd
        self.units = units
        self.rho = rho

        # set name of the armour unit
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

    def __str__(self):
        table = []

        for i, (volume, dimensions) in enumerate(self.units.items()):
            if i == 0:
                # set headers of the table
                headers = list(dimensions.keys())
                headers.insert(0, 'V')
            row = []
            row.append(volume)

            for parameter, dimension in dimensions.items():
                row.append(dimension)
            table.append(row)

        table = tabulate(table, headers, tablefmt='github')
        return str(table)

    def __setitem__(self, key, value):
        self.units[key] = value

    def __getitem__(self, key):
        return self.units[key]

    def get_class(self, d):
        """ Get the volume of a standard concrete armour unit

        Parameters
        ----------
        d : float
            minimal required diameter [m]

        Returns
        -------
        float
            Volume of a standard armour unit

        Raises
        ------
        ArmourUnitsError
            If the computed Dn of the armour layer is out of range for
            the specified armour units
        """
        computed_volume = d**3
        class_volume = None

        for volume in self.units:
            if volume >= computed_volume:
                class_volume = volume
                break

        if class_volume == None:
            max_class = list(self.units.keys())[-1]
            max_dn = np.round(max_class**(1/3), 3)
            msg = ('given dn is out of range for the specified armour units'
                   f', {max_dn} m (V = {max_class} m^3) is the maximum '
                   'possible dn.')
            raise ArmourUnitsError(msg)

        return class_volume


class Xbloc(ConcreteArmour):
    """ Xbloc concrete armour units

    .. figure:: _figures/Xbloc.png
       :align: center
       :alt: Xbloc armour unit

    Xbloc is an armour unit developed by Delta Marine Consultants (DMC).
    All units have been predefined in :py:attr:`units`, with the values
    from table 1 from Delta Marine Consultants (2018).

    Parameters
    ----------
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units

    Attributes
    ----------
    kd : int
        Stability coefficient [-]
    units : dict
        Dictionary of the defined concrete armour units. Format is
        {V: dict}, where V is the volume of the Xbloc (float) and
        dict = {'D': float, 'h': float, 'Vc': float} in
        which D is the diameter, h is the thickness of the armour
        layers and Vc is the volume of concrete.
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units [kg/m³]
    """

    def __init__(self, rho=2400):
        """ See help(Xbloc) for more info """
        units = {0.75: {'D': 1.31, 'h': 1.3, 'Vc': 0.53},
                 1: {'D': 1.44, 'h': 1.4, 'Vc': 0.58},
                 1.5: {'D': 1.65, 'h': 1.6, 'Vc': 0.66},
                 2: {'D': 1.82, 'h': 1.8, 'Vc': 0.73},
                 2.5: {'D': 1.96, 'h': 1.9, 'Vc': 0.78},
                 3: {'D': 2.08, 'h': 2, 'Vc': 0.83},
                 4: {'D': 2.29, 'h': 2.2, 'Vc': 0.92},
                 5: {'D': 2.47, 'h': 2.4, 'Vc': 0.99},
                 6: {'D': 2.62, 'h': 2.5, 'Vc': 1.05},
                 7: {'D': 2.76, 'h': 2.7, 'Vc': 1.11},
                 8: {'D': 2.88, 'h': 2.8, 'Vc': 1.16},
                 9: {'D': 3, 'h': 2.9, 'Vc': 1.2},
                 10: {'D': 3.11, 'h': 3, 'Vc': 1.25},
                 12: {'D': 3.3, 'h': 3.2, 'Vc': 1.32},
                 14: {'D': 3.48, 'h': 3.4, 'Vc': 1.39},
                 16: {'D': 3.63, 'h': 3.5, 'Vc': 1.46},
                 18: {'D': 3.78, 'h': 3.7, 'Vc': 1.52},
                 20: {'D': 3.91, 'h': 3.8, 'Vc': 1.57}}
        super().__init__(kd=16, units=units, rho=rho)

    def correction_factor(
            self, Hs, h, Rc, occurrence_hs, slope, slope_foreshore,
            permeability, logger=None, **kwargs):
        """ Determine correction factor for Xbloc

        Correction factors for phenomena which require an increase in
        the Xbloc unit sizes. For the conceptual design of structures
        the correction factor must be multiplied with the volume
        (Delta Marine Consultants, 2018).

        Parameters
        ----------
        Hs : float
            significant wave height [m]
        h : float
            water depth [m]
        Rc : float
            Crest freeboard [m]
        occurrence_hs : bool
            True if frequent occurrence of the near-design wave height
            during the lifetime of the structure. False if no frequent
            occurrence
        slope : float
            slope of the armour layer [rad]
        slope_foreshore : float
            slope of the foreshore [rad]
        permeability : {'permeable', 'low', 'impermeable'}
            Permeability of the core
        logger : dict, optional, default: None
            dict to log messages, must have keys 'INFO' and 'WARNINGS'

        Returns
        -------
        float
            The correction factor to be applied on the volume. Returns
            1 if no correction factor is needed.
        """
        correction = [1]
        log = [None]

        # frequent occurrence of near-design wave height
        if occurrence_hs:
            correction.append(1.25)
            log.append('Xbloc: frequent occurrence of Hs')

        # steep foreshore
        if np.arctan(1/30) <= slope_foreshore:
            if np.arctan(1/20) <= slope_foreshore:
                if np.arctan(1/15) <= slope_foreshore:
                    if np.arctan(1/10) <= slope_foreshore:
                        correction.append(2) # steeper than 1:10
                        log.append('Xbloc: steepness of foreshore is greater '
                                   'than 1:10')
                    else:
                        correction.append(1.5) # between 1:15-1:10
                        log.append('Xbloc: steepness of foreshore is between '
                                   '1:15 and 1:10')
                else:
                    correction.append(1.25) # between 1:20-1:15
                    log.append('Xbloc: steepness of foreshore is between '
                               '1:20 and 1:15')
            else:
                correction.append(1.1) # between 1:30-1:20
                log.append('Xbloc: steepness of foreshore is between 1:30 and'
                           ' 1:20')

        # Check if structure is low crested
        if Rc/Hs < 1:
            if Rc/Hs < 0.5:
                correction.append(2)
                log.append('Xbloc: low crested structure, Rc/Hs < 0.5')
            else:
                correction.append(1.5)
                log.append('Xbloc: low crested structure, Rc/Hs < 1.0')

        # Check if water depth is large
        if h > 2.5*Hs:
            if h > 3.5*Hs:
                correction.append(2)
                log.append('Xbloc: large water depth, h > 3.5 x Hs')
            else:
                correction.append(1.5)
                log.append('Xbloc: large water depth, h > 2.5 x Hs')


        # low core permeability
        if permeability == 'low':
            correction.append(1.5)
            log.append('Xbloc: low core permeability')
        elif permeability == 'impermeable':
            correction.append(2)
            log.append('Xbloc: impermeable core')

        # Check if armour slope is mild
        if slope < np.arctan(2/3):
            if slope < np.arctan(1/2):
                correction.append(1.5)
                log.append('Xbloc: armour slope is milder than 1:2')
            else:
                correction.append(1.25)
                log.append('Xbloc: armour slope is milder than 2:3')

        max_correction = max(correction)
        max_index = correction.index(max_correction)
        msg = (f'Xbloc: used correction factor: {max_correction}, because '
               f'{log[max_index]} was the maximum value')

        if logger is not None:
            if len(log) > 1:
                correction.pop(0)
                log.pop(0)
                logger['INFO'].extend(log)
                logger['INFO'].append(msg)

        return max_correction


class XblocPlus(ConcreteArmour):
    """ XblocPlus concrete armour units

    .. figure:: _figures/XblocPlus.png
       :align: center
       :alt: XblocPlus armour unit

    XblocPlus is an armour unit developed by Delta Marine Consultants
    (DMC). All units have been predefined in :py:attr:`units`, with the
    values from table 2 from Delta Marine Consultants (2018).

    Parameters
    ----------
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units

    Attributes
    ----------
    kd : int
        Stability coefficient [-]
    units : dict
        Dictionary of the defined concrete armour units. Format is
        {V: dict}, where V is the volume of the XblocPlus unit (float)
        and dict = {'L1': float, 'L2': float, 'L3': float, 'h': float
        'Vc': float} in which L1, L2 and L3 are the dimensions, h is
        the thickness of the armour layers and Vc is the volume of
        concrete.
    rho : float, optional, default: 2400
        density of the concrete used to make the armour units [kg/m³]
    """
    def __init__(self, rho=2400):
        """ See help(XblocPlus) for more info """
        units = {0.75: {'L1': 0.75, 'L2': 1.51, 'L3': 1.91, 'h': 1.2, 'Vc': 0.48},
                 1: {'L1': 0.83, 'L2': 1.66, 'L3': 2.1, 'h': 1.3, 'Vc': 0.53},
                 1.5: {'L1': 0.95, 'L2': 1.9, 'L3': 2.41, 'h': 1.5, 'Vc': 0.6},
                 2: {'L1': 1.04, 'L2': 2.09, 'L3': 2.65, 'h': 1.7, 'Vc': 0.66},
                 2.5: {'L1': 1.12, 'L2': 2.25, 'L3': 2.85, 'h': 1.8, 'Vc': 0.71},
                 3: {'L1': 1.19, 'L2': 2.39, 'L3': 3.03, 'h': 1.9, 'Vc': 0.76},
                 4: {'L1': 1.31, 'L2': 2.63, 'L3': 3.34, 'h': 2.1, 'Vc': 0.83},
                 5: {'L1': 1.42, 'L2': 2.83, 'L3': 3.59, 'h': 2.3, 'Vc': 0.9},
                 6: {'L1': 1.5, 'L2': 3.01, 'L3': 3.82, 'h': 2.4, 'Vc': 0.96},
                 7: {'L1': 1.58, 'L2': 3.17, 'L3': 4.02, 'h': 2.5, 'Vc': 1.01},
                 8: {'L1': 1.66, 'L2': 3.31, 'L3': 4.2, 'h': 2.7, 'Vc': 1.05},
                 9: {'L1': 1.72, 'L2': 3.45, 'L3': 4.37, 'h': 2.8, 'Vc': 1.09},
                 10: {'L1': 1.78, 'L2': 3.57, 'L3': 4.53, 'h': 2.9, 'Vc': 1.13},
                 12: {'L1': 1.89, 'L2': 3.79, 'L3': 4.81, 'h': 3.0, 'Vc': 1.2},
                 14: {'L1': 1.99, 'L2': 3.99, 'L3': 5.06, 'h': 3.2, 'Vc': 1.27},
                 16: {'L1': 2.09, 'L2': 4.17, 'L3': 5.29, 'h': 3.3, 'Vc': 1.33},
                 18: {'L1': 2.17, 'L2': 4.34, 'L3': 5.5, 'h': 3.5, 'Vc': 1.38},
                 20: {'L1': 2.65, 'L2': 4.5, 'L3': 5.7, 'h': 3.6, 'Vc': 1.43}}
        super().__init__(kd=12, units=units, rho=rho)

    def correction_factor(
            self, Hs, Rc, slope_foreshore, permeability, logger=None,
            **kwargs):
        """ Determine correction factor for XblocPlus

        Correction factors for phenomena which require an increase in
        the XblocPlus unit sizes. For the conceptual design of
        structures the correction factor must be multiplied with the
        volume (Delta Marine Consultants, 2018).

        Parameters
        ----------
        Hs : float
            significant wave height [m]
        Rc : float
            Crest freeboard [m]
        slope_foreshore : float
            slope of the foreshore [rad]
        permeability : str
            Permeability of the core {'permeable', 'low', 'impermeable'}
        logger : dict, optional, default: None
            dict to log messages, must have keys 'INFO' and 'WARNINGS'

        Returns
        -------
        float
            The correction factor to be applied on the volume. Returns
            1 if no correction factor is needed.
        """
        correction = [1]
        log = [None]

        # steep foreshore
        if np.arctan(1/30) <= slope_foreshore:
            if np.arctan(1/20) <= slope_foreshore:
                if np.arctan(1/15) <= slope_foreshore:
                    if np.arctan(1/10) <= slope_foreshore:
                        correction.append(2) # steeper than 1:10
                        log.append('XblocPlus: steepness of foreshore is '
                                   'greater than 1:10')
                    else:
                        correction.append(1.5) # between 1:15-1:10
                        log.append('XblocPlus: steepness of foreshore is '
                                   'between 1:15 and 1:10')
                else:
                    correction.append(1.25) # between 1:20-1:15
                    log.append('XblocPlus: steepness of foreshore is between '
                               '1:20 and 1:15')
            else:
                correction.append(1.1) # between 1:30-1:20
                log.append('XblocPlus: steepness of foreshore is between 1:30'
                           ' and 1:20')

        # Check if structure is low crested
        if Rc/Hs < 1:
            if Rc/Hs < 0.5:
                correction.append(1.5)
                log.append('XblocPlus: low crested structure, Rc/Hs < 0.5')
            else:
                correction.append(1.25)
                log.append('XblocPlus: low crested structure, Rc/Hs < 1.0')

        # low core permeability
        if permeability == 'low':
            correction.append(1.25)
            log.append('XblocPlus: low core permeability')
        elif permeability == 'impermeable':
            correction.append(1.5)
            log.append('XblocPlus: impermeable core')

        max_correction = max(correction)
        max_index = correction.index(max_correction)
        msg = (f'XblocPlus: used correction factor: {max_correction}, because '
               f'{log[max_index]} was the maximum value')

        if logger is not None:
            if len(log) > 1:
                correction.pop(0)
                log.pop(0)
                logger['INFO'].extend(log)
                logger['INFO'].append(msg)

        return max_correction
